"""
Many of the functions here defined are originally published by Alain de Cheveigné
http://audition.ens.fr/adc/NoiseTools/
"""
import numpy as np
import pyfftw
from tqdm import tqdm
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
import numba
__author__ = 'jaime undurraga'


def et_covariance(x: np.array = np.array([]),
                  shifts: np.array(int) = np.array([0]),
                  w: np.array = np.array([])) -> (np.array, float):
    """
    Compute covariance matrix

    :param x: numpy array with data
    :param shifts: integer array specifying shifts
    :param w: numpy array with weights
    :return c: numpy array with covariance matrix
    :return tw: weights norm
    """
    c = np.zeros((x.shape[1], x.shape[1]))
    if x.ndim == 2:
        x = x[:, :, None]
    if w.size == 0:
        for i in np.arange(x.shape[2]):
            xx = et_multi_shift(x[:, :, i], shifts)
            xx = xx.squeeze()
            c = c + xx.T.dot(xx)
        tw = xx.shape[0] * x.shape[2]
    else:
        tw = 0
        for i in np.arange(x.shape[2]):
            xx = x[:, :, i]
            xx = et_multi_shift(xx, shifts)
            ww = w[:, :, i]
            ww = et_multi_shift(ww, shifts)
            xx = xx * ww
            xx = xx.squeeze()
            c = c + xx.T.dot(xx)
            tw = tw + ww.T.dot(ww)
        # tw = np.sum(w[:])
        tw = xx.shape[0] * x.shape[2]
        # tw = np.sum(w[:, 0, :])
    return c, tw


def et_time_shifted_xcovariance(x: np.array = np.array([]),
                                y: np.array = np.array([]),
                                shifts: np.array(int) = np.array([0]),
                                wx: np.array = np.array([]),
                                wy: np.array = np.array([])) -> (np.array, float):
    """
        Compute covariance matrix between x and y

        :param x: numpy array with data
        :param y: numpy array with data
        :param shifts: integer array specifying shifts
        :param wx: numpy array with weights for x
        :param wy: numpy array with weights for y
        :return c: numpy array with covariance matrix
        :return tw: weights norm
        """

    n_shifts = shifts.size
    c = np.zeros((x.shape[1], y.shape[1] * n_shifts))
    if x.ndim == 2:
        x = x[:, :, None]
    if y.ndim == 2:
        y = y[:, :, None]
    if not wx.size:
        wx = np.ones(x.shape)
    if not wy.size:
        wy = np.ones(y.shape)
    # w_x = w_x / np.sum(w_x, axis=2, keepdims=True)
    # w_y = w_y / np.sum(w_y, axis=2, keepdims=True)
    tw = 0
    for i in np.arange(x.shape[2]):
        yy = et_multi_shift(y[:, :, i], shifts)
        wwy = et_multi_shift(wy[:, :, i], shifts)
        xx = x[0: yy.shape[0], :, i]
        wxx = wx[0: yy.shape[0], :, i]
        c = c + (xx * wxx).T.dot(yy * wwy)
        w_total = wxx.T.dot(wwy)
        tw = tw + w_total

    # for i in np.arange(x.shape[2]):
    #     yy = et_multi_shift(y[:, :, i], shifts)
    #     wwy = wy[:, :, i]
    #     xx = x[0: yy.shape[0], :, i]
    #     # wwx = wx[0: yy.shape[0], :, i]
    #
    #     c = c + xx * wwx.T.dot(yy * wwy)
    #
    # if not wy.size:
    #     tw = xx.shape[0] * y.shape[1] * yy.shape[0]
    # else:
    #     wy = wy[0: yy.shape[0], :, :]
    #     tw = np.sum(wy[:])
    return c, tw


def et_freq_shifted_xcovariance(x: np.array = np.array([]),
                                y: np.array = np.array([]),
                                shifts: np.array(int) = np.array([0]),
                                wx: np.array = np.array([]),
                                wy: np.array = np.array([]),
                                normalized_frequencies: np.array = np.array([]),
                                n_jobs: int = 1) -> (np.array, float):
    """
        Compute covariance matrix between x and y in the frequency domain
        :param x: numpy array with data
        :param y: numpy array with data
        :param shifts: integer array specifying shifts
        :param wx: numpy array with weights to be applied to x
        :param wy: numpy array with weights to be applied to y
        :param normalized_frequencies: normalized freqs to compute covariance
        :param n_jobs: number of CPUs to compute FFT
        :return c: numpy array with covariance matrix
        :return tw: weights norm
        """

    n_shifts = shifts.size
    c = np.zeros((x.shape[1], y.shape[1] * n_shifts))
    if x.ndim == 2:
        x = x[:, :, None]
    if y.ndim == 2:
        y = y[:, :, None]
    if not wx.size:
        wx = np.ones(x.shape)
    if not wy.size:
        wy = np.ones(y.shape)

    if x.ndim == 2:
        x = x[:, :, None]
    if np.mod(x.shape[0], 2) == 0:
        fft_size = int((x.shape[0] / 2) + 1)
    else:
        fft_size = int((x.shape[0] + 1) / 2)

    freq = np.arange(fft_size) / float(x.shape[0])
    freq_filt = np.zeros(shape=(fft_size, 1), dtype=float)
    for f_bin in et_find_freq_bin(freq, normalized_frequencies):
        freq_filt[f_bin] = 1.0

    c = 0
    tw = 0
    for i in np.arange(x.shape[2]):
        yy = et_multi_shift(y[:, :, i], shifts)
        wyy = et_multi_shift(wy[:, :, i], shifts)
        xx = x[0: yy.shape[0], :, i]
        wxx = wx[0: yy.shape[0], :, i]
        xx = xx * wxx
        yy = y[:, :, i] * wyy
        xfft = pyfftw.builders.rfft(xx,
                                    overwrite_input=False,
                                    planner_effort='FFTW_ESTIMATE',
                                    axis=0,
                                    threads=n_jobs)()
        yfft = pyfftw.builders.rfft(yy,
                                    overwrite_input=False,
                                    planner_effort='FFTW_ESTIMATE',
                                    axis=0,
                                    threads=n_jobs)()
        zx = xfft * freq_filt
        zy = yfft * freq_filt
        c = c + 2 * np.real(zx.T.conj().dot(zy))
        tw = tw + wxx.T.dot(wyy)

    return c, tw


def et_unfold(x, orthogonal=False):
    # unfold to improve machine performance during multiplications
    if orthogonal:
        return np.reshape(np.transpose(x, [0, 2, 1]), [-1, x.shape[1]], order='F')
    else:
        return np.reshape(np.transpose(x, [0, 2, 1]), [-1, x.shape[1]])


def et_fold(x, epoch_size):
    #  fold back to normal
    return np.transpose(np.reshape(x, [epoch_size, -1, x.shape[1]]), [0, 2, 1])


def et_mmat(x, m):
    return et_fold(et_unfold(x).dot(m), x.shape[0])


def et_x_covariance(x, y):
    c = np.zeros((x.shape[1], y.shape[1]))
    for i in np.arange(x.shape[2]):
        c += x[:, :, i].T.dot(y[:, :, i])
    return c


def et_weighted_covariance(
        x: np.array = np.array([]),
        w: np.array([]) = np.array([])) -> (np.array, np.array):
    """
    Compute the weighted covariance of x
    :param x: numpy array with data to compute weights
    :param w: numpy array with weights to be used
    :return: covariance matrix and the total weights
    """

    #
    #  X can be 1D, 2D or 3D.
    #  W can be 1D (if X is 1D or 2D) or 2D (if X is 3D). The same weight is
    #  applied to each column.
    if w.size == 0:
        w = np.ones(x.shape)
    assert x.shape[0] == w.shape[0], "x and w must have the same number of samples"

    if w.shape[1] == 1:
        w = np.tile(w, [1, x.shape[1]])
    assert w.shape[1] == x.shape[1], "x and w must have the same number of columns"
    if x.ndim == 2:
        x = x[:, :, None]
    if w.ndim == 2:
        w = w[:, :, None]
    # x = et_unfold(x)
    # w = et_unfold(w)
    #  weights
    tw = 0
    c = 0
    for i in np.arange(x.shape[2]):
        ww = w[:, :, i]
        xx = x[:, :, i] * ww
        w_total = ww.T.dot(ww)
        tw = tw + w_total
        c = c + xx.T.dot(xx)
    return c, tw


def et_x_join_covariance(x, y):
    # compute the covariance combining all epochs
    return et_unfold(x).T.dot(et_unfold(y))


@numba.jit(nogil=True, nopython=True)
def et_find_freq_bin(freq_vector: np.array = np.array([]),
                     freq_to_find: np.array = np.array([])):
    f_bins = []
    for f in freq_to_find:
        f_pos = np.argmin(np.abs(freq_vector - f))
        f_bins.append(f_pos)
    return f_bins


# def et_w_covariance(epochs: np.array = np.array([])):
#     c = np.zeros((epochs.shape[1], epochs.shape[1]))
    # w_mean, w = et_weighted_mean(epochs)
    # for i in np.arange(x.shape[2]):
    #     c = c + x[:, :, i].T.dot(x[:, :, i])
    # return cfol

def et_multi_shift(x: type(np.array) | None = None,
                   shifts: np.array(int) = np.array([0])):
    """
    :param  x: matrix to shift
    :param shifts: array of shifts (must be nonnegative)
    :return z: shifted matrix
    """
    z = x
    if shifts.size > 0 and np.any(shifts > 0):
        assert x.shape[0] > np.max(shifts), 'shifts should be no larger than nrows'
        assert np.min(shifts) >= 0, 'shifts should be nonnegative'
        n_shifts = shifts.shape[0]
        n = x.shape[0] - np.max(shifts)
        shift_array = (np.ones((n, n_shifts)) * shifts + np.arange(0, n).reshape(-1, 1)).astype(int)
        if x.ndim == 2:
            x = x[:, :, None]
        n_s, n_ch, n_trials = x.shape
        n_trials = np.maximum(n_trials, 1)
        z = np.zeros((n, n_ch * n_shifts, n_trials))
        for k in np.arange(n_trials):
            for j in np.arange(0, n_ch):
                y = x[:, j, k]
                pos = np.arange(j * n_shifts, j * n_shifts + n_shifts).astype(int)
                z[:, pos, k] = y[shift_array]
    z = z.squeeze()
    return z


def et_freq_weighted_cov(x: np.array = np.array([]),
                         normalized_frequencies: np.array = np.array([]),
                         w: np.array = np.array([]),
                         n_jobs: int = 1):
    if x.ndim == 2:
        x = x[:, :, None]
    if np.mod(x.shape[0], 2) == 0:
        fft_size = int((x.shape[0] / 2) + 1)
    else:
        fft_size = int((x.shape[0] + 1) / 2)

    freq = np.arange(fft_size) / (float(x.shape[0]))
    freq_filt = np.zeros(shape=(fft_size, 1), dtype=float)
    for f_bin in et_find_freq_bin(freq, normalized_frequencies):
        freq_filt[f_bin] = 1.0
    if w.size == 0:
        w = np.ones(x.shape)
        # w = w / np.sum(w, axis=2, keepdims=True)
    # normalize weights
    # w = w / np.sum(w ** 2, axis=2, keepdims=True)
    c = 0
    tw = 0
    xfft = pyfftw.builders.rfft(x * w,
                                overwrite_input=False,
                                planner_effort='FFTW_ESTIMATE',
                                axis=0,
                                threads=n_jobs)()

    for i in tqdm(np.arange(x.shape[2]), desc='Spatial filter covariance estimation'):
        ww = w[:, :, i]
        # xx = x[:, :, i] * ww
        # fft = pyfftw.builders.rfft(xx, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
        #                            threads=multiprocessing.cpu_count())
        # xfft = fft()
        # xfft = np.fft.rfft(xx, axis=0)
        z = xfft[:, :, i] * freq_filt
        c = c + 2 * np.real(z.T.conj().dot(z))
        tw = tw + ww.T.dot(ww)
    return c, tw


# def et_tsr(x: np.array = np.array([]),
#            ref: np.array = np.array([]),
#            shifts: np.array(int) = np.array([0]),
#            wx: np.array(int) = np.array([0]),
#            wref: np.array = np.array([]),
#            keep: np.array = np.array([]),
#            thresh: float = 1e-20) -> np.array:
#     """
#         Compute covariance matrix between x and y
#         :param x: numpy array with data
#         :param y: numpy array with data
#         :param shifts: integer array specifying shifts
#         :param wx: numpy array with weights
#         :param wy: numpy array with weights
#         :return c: numpy array with covariance matrix
#         :return tw: weights norm
#         """
#     # We need to adjust x and ref to ensure that shifts are non-negative.
#     # If some values of shifts are negative, we increment shifts and truncate x.
#     # adjust x to make shifts non-negative
#     offset1 = np.maximum(0, -shifts.min())
#     idx = np.arange(offset1, x.shape[0])
#     x = x[idx, :, :]  # truncate x
#     if wx.size > 0:
#         wx = wx[idx, :, :]  # truncate wx
#     shifts = shifts + offset1  # shifts are now positive
#     # adjust size of x
#     offset2 = np.maximum(0, shifts.max())
#     idx = np.arange(0, ref.shape[0] - offset2)
#     x = x[idx, :, :]  # part of x that overlaps with time-shifted refs
#     if wx.size > 0:
#         wx = wx[idx, :, :]  # truncate wx
#     [mx,nx,ox] = x.shape
#     [mref, nref, oref] = ref.shape
#     # consolidate weights into single weight matrix
#     w = np.zeros([mx, 1, oref])
#     if wx.size == 0 and wref.size == 0:
#         w[0: mx, :, :] = 1
#     elif wref.size == 0:
#         w[:, :, :] = wx[:, :, :]
#     elif wx.size == 0:
#         for k in range(ox):
#             wr = wref[:, :, k]
#             wr = ts_multishift(wr, shifts)
#             wr=np.min(wr, axis = 1)
#             w[:, :, k] = wr
#     else:
#         for k in range(ox):
#             wr = wref[:, :, k]
#             wr = nt_multishift(wr, shifts)
#             wr = np.min(wr, axis=1)
#             wr = np.min(wr, wx[0: wr.shape[0], :, k])
#             w[:, :, k] = wr
#     wx = w
#     wref = np.zeros(mref, 1, oref)
#     wref[idx, :, :] = w
#     # remove weighted means
#     x0 = x
#     x = et_demean(x, wx)
#     mn1 = x - x0
#     ref = et_demean(ref,wref)
#     # equalize power of ref chans, then equalize power of ref PCs
#     ref = nt_normcol(ref,wref)
#     ref = nt_pca(ref, 0, [], 1e-6)
#     ref = nt_normcol(ref, wref)
#     # covariances and cross covariance with time-shifted refs
#     [cref,twcref] = nt_cov(ref, shifts, wref)
#     [cxref,twcxref] = nt_xcov(x, ref, shifts, wx)
#     # regression matrix of x on time-shifted refs
#     r = nt_regcov(cxref/twcxref,cref/twcref,keep,thresh)
#     #  r=r*0.765;
#     #  TSPCA: clean x by removing regression on time-shifted refs
#     y = np.zeros[mx, nx, ox]
#     for k in range(ox):
#         z = nt_multishift(ref[:, :, k], shifts) * r
#         y[:, :, k] = x[0:z.shape[0], : , k] - z
#     y0 = y
#     y = et_demean(y, wx)    # multishift(ref) is not necessarily 0 mean
#     mn2 = y - y0
#     idx = np.arange(offset1, y.shape[0] + offset1)
#     mn = mn1 + mn2
#     w = wref
#     return y
#
# def nt_tsregress(x: np.array = np.array([]),
#                  y: np.array = np.array([]),
#                  shifts: np.array(int) = np.array([0]),
#                  wx: np.array(int) = np.array([0]),
#                  wy: np.array(int) = np.array([0]),
#                  keep: np.array = np.array([]),
#                  threshold: float = 1e-20) -> np.array:
#
#     #  z: part of x modeled by time-shifted y
#     #  idx: x(idx) maps to z
#
#     #  x: data to model
#     #  y: regressor
#     #  shifts: array of shifts to apply (default: [0])
#     #  xw: weights to apply to x
#     #  yw: weights to apply to y
#     #  keep: number of components of shifted regressor PCs to keep (default: all)
#     #  threshold: discard PCs with eigenvalues below this
#
#     # Data X are regressed on time-shifted versions of Y. X and Y are initially
#     # time-aligned, but because of the shifts, Z is shorter than X.  Z is
#     # time-aligned with X(IDX).
#
#     assert x.shape[0] == y.shape[0], "x and y must have same number of samples"
#
#     # shifts must be non-negative
#     mn = shifts.min()
#     if mn < 0:
#         shifts = shifts - mn
#         x = x[-mn::, :, :]
#         y = y[-mn::, :, :]
#
#     nshifts = shifts.size
#
#     #  #  flag outliers in x and y
#     #  if ~isempty(toobig1) || ~isempty(toobig2)
#     #      xw=nt_find_outliers(x,toobig1,toobig2);
#     #      yw=nt_find_outliers(y,toobig1,toobig2);
#     #  else
#     #      xw=[];yw=[];
#     #      # xw=ones(size(x)); yw=ones(size(y));
#     #  end
#
#     #  subtract weighted means
#
#     if x.ndim == 3:
#         [Mx, Nx, Ox] = x.shape
#         [My, Ny, Oy] = y.shape
#         x=et_unfold(x)
#         y=et_unfold(y)
#         [x, xmn] = et_demean(x, wx)
#         [y, ymn] = et_demean(y, wy)
#         x = et_fold(x, Mx)
#         y = et_fold(y, My)
#     else:
#         [x, xmn] = et_demean(x, wx)
#         [y, ymn] = et_demean(y, wy)
#
#     #  covariance of y
#     cyy, totalweight = et_covariance(y, shifts.T, wy)
#     cyy = cyy / totalweight
#
#     #  cross-covariance of x and y
#     [cxy, totalweight] = nt_cov2(x, y, shifts.T, wx, wy)
#     #disp('!!!!!!!!!   WARNING: calling obsolete code  !!!!!!!!!!!!!!!!');
#     # [cxy, totalweight]=nt_xcov(x,y,shifts',xw,yw);
#     cxy= cxy / totalweight
#
#     #  regression matrix
#     r = nt_regcov(cxy, cyy, keep, threshold)
#
#     #  regression
#     if x.ndim == 3:
#         x = et_unfold(x)
#         y = et_unfold(y)
#         [m, n, l] = x.shape
#         mm = m - shifts.max()
#         z = np.zeros(x.shape)
#         for k in range(nshifts):
#             kk = shifts(k)
#             idx1 = np.arange(kk, kk + mm)
#             idx2 = k + np.arange(0, y.shape[1] - 1) * nshifts
#             z[0: mm, :] = z[0: mm, :] + y[idx1, :] * r[idx2, :]
#
#         z=et_fold(z, Mx)
#         z = z[0:-shifts.max(), :, :]
#     else:
#         [m, n] = x.shape
#         z = np.zeros(m - shifts.max(), n)
#         for k in range(nshifts):
#             kk = shifts[k]
#             idx1 = np.arange(kk, kk + z.shape[0])
#             idx2 = k + np.arange(0, y.shape[1] - 1) * nshifts
#             z = z + y[idx1, :] * r[idx2, :]
#
#     #  idx allows x to be aligned with z
#     offset = np.maximum(0, -mn)
#     idx=np.arange(offset, offset + z.shape[0])
#
#
#
def et_demean(x: np.array = np.array([]),
              w: np.array = np.array([])):

    if w.size and w.size < x.shape[0]:
        w = w[:]
        # interpret w as array of indices to set to 1
        assert ~(np.min(w) < 0 or np.max(w) > x.shape[0]), 'w interpreted as indices but values are out of range'
        ww = np.zeros((x.shape[0], 1, 1))
        ww[w] = 1
        w = ww

    if w.ndim == 3 and w.shape[2] != x.shape[2]:
        if w.shape[2] == 1 and x.shape[2] != 1:
            w = np.tile(w, [1, 1, x.shape[2]])
        else:
            raise Exception('W should have same number of trials as X, or else 1')

    [m, n, o] = x.shape
    # xx = et_unfold(x)
    # ww = et_unfold(w)
    # mm = xx * ww / np.sum(ww, axis=0)
    # xx = xx - mm
    # x = et_fold(xx, x.shape[0])

    if w.size == 0:
        mn = np.mean(x, axis=0)
        x = x - mn
    else:
        # w = et_unfold(w)
        if w.shape[0] != x.shape[0]:
            raise Exception('X and W should have same number of rows')
        if w.shape[1] == 1:
            mn = np.sum(x * w, axis=0) / (np.sum(w, axis=0))
        elif w.shape[1] == n:
            mn = np.sum(x * w, axis=0) / (np.sum(w, axis=0))
        else:
            raise Exception('W should have same number of columns as X, or else 1')
        x = x - mn
    return x


def w_mean(epochs: np.array = np.array([]),
           weights: np.array = np.array([]),
           keepdims=False,
           axis=2):
    """
    Computes weighted mean using provided weights
    :param epochs: samples x channels x trials numpy array
    :param weights: samples x channels x trials numpy array or channels x trials numpy (assumed same weights across
    samples). If samples == 1, then same weight is use across samples
    :param keepdims: if True, dimension of original input is kept
    :param axis: direction in which to average
    :return: weighted average
    """
    if weights.size == 0:
        weights = np.ones(epochs.shape)
    if weights.ndim == 2 and epochs.ndim == 3:
        assert epochs.shape[1::] == weights.shape, "number of channels and trials must agree between epochs and weights"
    if weights.ndim == 3:
        if weights.shape[0] != 1:
            assert epochs.shape == weights.shape, "number samples, trials, and channels must agree between epochs and" \
                                                  " weights"
    if weights.ndim == 2:
        weights = np.tile(weights, (epochs.shape[0], 1, 1))
    # normalize weights across trials
    weights = weights / np.expand_dims(np.sum(weights, axis=axis), axis=axis)
    return np.sum(epochs * weights, axis=axis, keepdims=keepdims)


def effective_sampling_size(weights: np.array, axis=2):
    """
    compute the effective sample size resulting from the use of weights
    Kish, Leslie(1965). "Survey Sampling".New York: John  Wiley & Sons, Inc.ISBN 0 - 471 - 10949 - 5.
    "Design Effects and Effective Sample Size".
    https://docs.displayr.com/wiki/Design_Effects_and_Effective_Sample_Size

    :param weights: array of weights
    :param axis: direction indicating how weights are summed
    :return:
    """
    weights = set_default_unit(weights, u.dimensionless_unscaled)
    ess = np.sum(weights, axis=axis) ** 2.0 / (np.sum(weights ** 2.0, axis=axis))
    return ess



import peegy.io.tools.xml_tools as x_tools
import numpy as np
import os
from os import path
import json
import zipfile as zip_f
import glob
import tempfile
from pathlib import Path
import shutil
from typing import List


def get_measurement_info_from_zip(zip_file_name):
    with zip_f.ZipFile(zip_file_name, 'r') as my_zip:
        files = my_zip.infolist()
        m_file = my_zip.open(files[0].filename, 'r')
        _, _extension = path.splitext(files[0].filename)
        if _extension == '.xml':
            content = m_file.read()
            measurement_info = x_tools.xml_string_to_dict(content)
        if _extension == '.json':
            content = m_file.readlines()
            measurement_info = json.loads(str.replace(content[0].decode("utf-8"), 'Inf', '''100000'''))
        return measurement_info


def get_extsys_parameters(parameter_files=['']):
    parameters = []
    for i, _file_name in enumerate(parameter_files):
        parameters.append(get_measurement_info_from_zip(_file_name))
    dates = [par['Measurement']['MeasurementModule']['Date'] for x, par in enumerate(parameters)]
    idx_par = [i[0] for i in sorted(enumerate(dates), key=lambda xx: xx[1])]
    sorted_parameters = np.array(parameters)[idx_par]
    sorted_parameter_files = np.array(parameter_files)[idx_par]
    return sorted_parameters, sorted_parameter_files, dates, idx_par


def freeze(d):
    if isinstance(d, dict):
        return frozenset((key, freeze(value)) for key, value in list(d.items()))
    elif isinstance(d, list):
        return tuple(freeze(value) for value in d)
    return d


def merge_parameters_by_condition(parameters=[{}], **kwargs):
    merge_conditions_by = kwargs.get('merge_conditions_by', [''])
    merge_stimulus_index = kwargs.get('merge_stimulus_index', 0)

    for _p in parameters:
        assert merge_stimulus_index < len(_p['Measurement']['StimuliModule']['Stimulus'])
        _par = _p['Measurement']['StimuliModule']['Stimulus'][merge_stimulus_index]['Parameters']

    _m_items = []
    for i_p, _p in enumerate(parameters):
        _par = _p['Measurement']['StimuliModule']['Stimulus'][merge_stimulus_index]['Parameters']
        _temp = []
        for _cond in merge_conditions_by:
            _temp.append(_par[_cond])
        _m_items.append(_temp)

    (conditions, index, inverse, counts) = np.unique(_m_items,
                                                     axis=0,
                                                     return_counts=True,
                                                     return_inverse=True,
                                                     return_index=True)
    sorted_parameters = parameters[index]
    return sorted_parameters, index, inverse


def anonymize(directory: str | None = None,
              subjects_original_id: List[str] | None = None,
              subjects_new_codes: List[str] | None = None):
    """
    This function will replace the original subject id in each .extsys file found within the directory and
    subdirectories. The old files will be renamed with their own name + '_old'.
    :param directory: root directory where .extsys files will be searched from.
    :param subjects_original_id: list of existing IDs
    :param subjects_new_codes: list of new codes use to replace the exiting IDs
    :return: None
    """
    assert len(subjects_original_id) == len(subjects_new_codes)
    assert np.unique(subjects_original_id).size == np.unique(subjects_new_codes).size
    files = sorted(glob.glob(directory + '**/*.extsys', recursive=True))
    for _file in files:
        zip_file_name = _file
        with zip_f.ZipFile(zip_file_name, 'r') as my_zip:
            files = my_zip.infolist()
            with tempfile.TemporaryDirectory() as tmpdirname:
                my_zip.extract(files[0].filename, path=tmpdirname)
                _meta_file = tmpdirname + os.sep + files[0].filename
                try:
                    with open(_meta_file, 'r') as _f:
                        measurement_info = json.load(_f)
                except json.JSONDecodeError:
                    continue
        _subject = measurement_info['Measurement']['MeasurementModule']['Subject']
        _idx_subject = np.argwhere(np.array(subjects_original_id) == _subject)
        if _idx_subject.size:
            _subject_code = subjects_new_codes[int(_idx_subject.squeeze())]
            measurement_info['Measurement']['MeasurementModule']['Subject'] = _subject_code
            os.rename(zip_file_name, zip_file_name + '_old')
            _file_path, _file_name = path.split(zip_file_name)
            _new_json_file = _file_path + os.sep + files[0].filename
            with open(_new_json_file, "x") as f:
                json.dump(measurement_info, f)
                f.close()
            new_zip_file_name = _file_path + os.path.sep + _file_name.replace(_subject, _subject_code)
            zipObj = zip_f.ZipFile(new_zip_file_name, 'w')
            zipObj.write(filename=_new_json_file, arcname=files[0].filename)
            zipObj.close()
            os.remove(_new_json_file)
        else:
            print('Subject {:} not found in subject codes'.format(_subject))


def extsys_to_json(input_file_name: str | None = None,
                   output_path: str | None = None,
                   output_file_name: str | None = None,
                   trash_original: bool = False):
    """
    This function will extract .json files from .extsys files.
    :param input_file_name: path to .extsys file
    :param output_path: directory where to .json file. If not given, file will be saved in the same directory as the
    .extsys file
    :param output_file_name: name of .json file. If empty, input_file_name will be used
    :param trash_original: If True, the input file will be moved to a .trash folder in the same directory
    :return:
    """

    if output_file_name is None:
        _, _file_name = os.path.split(input_file_name)
        output_file_name = _file_name

    if output_path is None:
        _path, _ = os.path.split(input_file_name)
        output_path = Path(_path)

    zip_file_name = input_file_name
    _, _extension = os.path.splitext(zip_file_name)
    assert _extension == '.extsys'

    with zip_f.ZipFile(zip_file_name, 'r') as my_zip:
        files = my_zip.infolist()
        with tempfile.TemporaryDirectory() as tmpdirname:
            my_zip.extract(files[0].filename, path=tmpdirname)
            _meta_file = tmpdirname + os.sep + files[0].filename
            try:
                with open(_meta_file, 'br') as _f:
                    _, _extension = path.splitext(_meta_file)
                    if _extension == '.xml':
                        content = _f.read()
                        measurement_info = x_tools.xml_string_to_dict(content)
                    if _extension == '.json':
                        measurement_info = json.load(_f)
            except json.JSONDecodeError:
                return
    _file_path, _file_name = path.split(zip_file_name)
    _output_file_name, _ = os.path.splitext(output_file_name)
    _new_json_file = output_path.joinpath(_output_file_name + '.json')
    with open(_new_json_file, "x") as f:
        json.dump(measurement_info, f)
        f.close()
    if trash_original:
        _path, _name = os.path.split(input_file_name)
        _trash_path = Path(_path).joinpath('.trash')
        Path(_trash_path).mkdir(parents=True, exist_ok=True)
        shutil.move(input_file_name, _trash_path.joinpath(_name))
        print('Original file {:} moved to {:}'.format(input_file_name,
                                                      _trash_path.joinpath(_name)))



from peegy.processing.tools.epochs_processing_tools import et_subtract_oeg_template
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
import numpy as np
from scipy.io import loadmat
import os
import astropy.units as u
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')


_directory = os.path.dirname(os.path.abspath(__file__))
data_path = _directory + os.sep + 'test_data' + os.sep + 'eog_blinks_valderrama_et_al_2018.mat'
data_structure = loadmat(data_path)
fs = float(data_structure['fs'].squeeze()) * u.Hz
data = data_structure['y'] * u.uV

# remove artefacts
clean_data, fig_results, template, z, eog_events = et_subtract_oeg_template(data=data,
                                                                            idx_ref=np.array([0]),
                                                                            fs=fs,
                                                                            n_iterations=15,
                                                                            plot_results=True,
                                                                            high_pass=None,
                                                                            low_pass=20 * u.Hz,
                                                                            template_width=1.4 * u.s,
                                                                            kernel_bandwidth=0.15,
                                                                            use_initial_template=True
                                                                            )
fig_results[0].show()

# SNR estimate
length = template.size
norm_2 = filt_data(data ** 2, np.ones(template.shape[0]), onset_padding=False, mode='valid')
z_norm_2 = z ** 2 / norm_2
z_norm_k2 = z_norm_2[eog_events + length - 1].value
snr = 10 * np.log10(np.mean(z_norm_k2 / (1 - z_norm_k2)))
print('SNR: {:}'.format(np.round(snr, decimals=2)))



import configparser
import re
import numpy as np
import peegy.definitions.eegReaderAbstractClasses as eegAbstract
__author__ = 'jundurraga'


class SynergyReader(eegAbstract.EEGData):
    def __init__(self, file_name=''):
        self.__file_name = file_name
        self.__data = np.array([])
        self.__channel_labels = []
        self.__channel_units = []
        self.__fs = []
        self.__gain = 1.0
        self.__time = np.array([])
        self.__channel_recording_parameters = []
        # initial parsing
        self.__parser = IniParser()
        self.__parser.read(file_name)
        self._file_data = self.__parser.as_dict()
        self.__parse_store_data()
        self.__time_offset = 0.0
        self.subject = dict(self._file_data['1 - Patient Details'], **self._file_data['1.1 - Visit'])

    def get_data(self):
        return self.__data / self.__gain

    def set_data(self, value):
        self.__data = value

    data = property(get_data, set_data)

    def get_gain(self):
        return self.__gain

    def set_gain(self, value):
        self.__gain = value

    gain = property(get_gain, set_gain)

    def set_channel_labels(self, value):
        self.__channel_labels = value

    def get_channel_labels(self):
        return self.__channel_labels

    channelLabels = property(get_channel_labels)

    def set_channel_units(self, value):
        self.__channel_units = value

    def get_channel_units(self):
        return self.__channel_units

    channel_units = property(get_channel_units, set_channel_units)

    def get_time(self):
        return np.arange(0, self.__data.shape[0]) / self.fs - self.__time_offset

    time = property(get_time)

    def set_time_offset(self, value):
        self.__time_offset = value

    def get_time_offset(self):
        return self.__time_offset

    time_offset = property(get_time_offset, set_time_offset)

    def get_channel_recording_parameters(self):
        return self.__channel_recording_parameters

    channel_recording_parameters = property(get_channel_recording_parameters)

    def time_to_samples(self, value):
        return np.where(self.time >= value)[0][0]

    def __parse_store_data(self):
        count = 0
        for _key_name in list(self._file_data.keys()):
            if re.search(r'(\s*)Store Data', _key_name):
                data_values_key = [_sub_key for _sub_key in self._file_data[_key_name] if re.search('averaged data',
                                                                                                    _sub_key)]
                _fs_key = [_sub_key for _sub_key in self._file_data[_key_name] if re.search('sampling frequency*',
                                                                                            _sub_key)][0]
                if not data_values_key:
                    continue
                data_unit = re.search(r'\([A-Za-z]*\)', data_values_key[0]).group()
                _data = np.expand_dims(
                    np.array(list(map(float, self._file_data[_key_name][data_values_key[0]].split(',')))),
                    axis=1)
                # self.__channel_labels.append(self._file_data[_key_name]['channel number'])
                self.__channel_labels.append(str(count))
                self.__channel_units.append(data_unit)
                self.__fs.append(float(self._file_data[_key_name][_fs_key]) * 1000.0)
                if not self.__data.any():
                    self.__data = np.append(self.__data, _data)
                else:
                    self.__data = np.hstack((self.__data, _data))
                if self.__data.ndim == 1:
                    self.__data = np.expand_dims(self.__data, axis=1)
                _temp_dict = self._file_data[_key_name].copy()
                _temp_dict.pop(data_values_key[0])
                self.__channel_recording_parameters.append(_temp_dict)
                count += 1

    def get_fs(self):
        return np.unique(self.__fs)

    def set_fs(self, new_value):
        self.__fs = new_value

    fs = property(get_fs, set_fs)

    def get_file_name(self):
        return self.__file_name

    def set_file_name(self, new_value):
        self.__file_name = new_value

    file_name = property(get_file_name, set_file_name)

    def get_epochs(self, win_length=0):
        return self.data

    def get_triggers(self, triggerCodes=[]):
        return []


class IniParser(configparser.ConfigParser):
    def as_dict(self):
        d = dict(self._sections)
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        return d



from peegy.io.paths.defaults import default_path
from peegy.io.eeg.reader import eeg_reader
from peegy.io.exporters.wav_writer import data_to_wav
from peegy.tools.signal_generator.noise_functions import generate_modulated_noise
import numpy as np
from peegy.processing.events.event_tools import Events, SingleEvent
import matplotlib.pyplot as plt
import astropy.units as u
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')


fs = 1000 * u.Hz
duration = 10 * u.s
n_channels = 1
data = generate_modulated_noise(
    fs=fs.to(u.Hz).value,
    duration=duration.to(u.s). value,
    n_channels=n_channels)
events = Events(events=np.array([SingleEvent(code=1,
                                             time_pos=_i * duration / 20,
                                             dur=1 * u.ms) for _i in range(10)]))

output_file_path = str(default_path.joinpath('test.wav'))
_file_path = ''
reader = eeg_reader(file_name=_file_path)

_wav_data, _wav_fs = data_to_wav(data=data,
                                 events=events,
                                 fs=fs,
                                 fs_wav=10000 * u.Hz,
                                 output_file_name=output_file_path)
#
fig1 = plt.figure(figsize=(7, 6), dpi=100)
t_original = np.arange(data.shape[0]) / fs
t_wav = np.arange(_wav_data.shape[0]) / _wav_fs
ax = fig1.add_subplot(1, 1, 1)
ax.plot(t_original, data)
ax.plot(t_wav, _wav_data)
plt.show()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import epochs_processing_tools as ept
import numpy as np
import matplotlib.pyplot as plt

# create synthetic data
fs = 4000.0
nsamples = np.round(1 * fs).astype(int)
nchans = 36
ntrials = 130
noise_dim = 36  # dimensionality of noise
f1 = 40
f2 = 8
t = np.arange(nsamples) / fs
source = np.expand_dims(np.sin(2 * np.pi * f1 * t) + np.sin(2 * np.pi * f2 * t), axis=1)
coeff = np.ones(nchans//2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)

s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))

desired_snr = 5.0
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * ntrials ** 0.5
theoretical_rn = ini_std / ntrials ** 0.5

# noise = np.random.normal(0, ini_std[0], size=(nsamples, nchans, ntrials))
noise = np.random.power(0.5, size=(nsamples, nchans, ntrials))
s[:, 0] = s[:, 0] * 0.5
data = noise + s


w_ave, w, rn, cumulative_rn, snr, cumulative_snr, s_var, w_fft, nk = ept.et_weighted_mean(epochs=data,
                                                                                          block_size=10,
                                                                                          samples_distance=10
                                                                                          )
across_channels_ave, total_rn, total_snr, t_s_var = \
    ept.et_snr_weighted_mean(averaged_epochs=w_ave, fs=fs, rn=rn, snr=np.max(snr, axis=0))

# fig = plt.figure()
# ax = fig.add_subplot(131)
# ax.plot(np.mean(s, axis=2))
# ax = fig.add_subplot(132)
# ax.plot(w_ave)
# ax = fig.add_subplot(133)
# ax.plot(across_channels_ave)
# plt.show()

freq = np.arange(0, w_fft.shape[0]) * fs / w_ave.shape[0]
# fig = plt.figure()
# ax = fig.add_subplot(111)
# ax.plot(freq, np.abs(w_fft))
# plt.show()

# moving averaging

data2 = data.transpose(2, 0, 1).reshape(-1, data.shape[1])
Nw = int(4 * 1. / f2 * fs)
k = data.shape[0] // Nw
new_epoch_number = k * data.shape[2] - k
new_epochs = np.zeros((data.shape[0], data.shape[1], new_epoch_number))
for _i in range(new_epoch_number):
    _ini_pos = int(round(_i * Nw))
    _end_pos = _ini_pos + data.shape[0]
    if _end_pos >= data2.shape[0]:
        break
    new_epochs[:, :, _i] = data2[_ini_pos:_end_pos, :]


w_ave_2, w_2, rn_2, cumulative_rn_2, snr_2, cumulative_snr_2, s_var_2, w_fft_2, nk_2 = \
    ept.et_weighted_mean(epochs=new_epochs,
                         block_size=10,
                         samples_distance=10)

fig = plt.figure()
ax = fig.add_subplot(131)
ax.plot(np.mean(s, axis=2))
ax = fig.add_subplot(132)
ax.plot(w_ave)
ax = fig.add_subplot(133)
ax.plot(w_ave_2)
plt.show()

# [f, px] = welch(data2, fs=fs, window='blackman', nperseg=4000., noverlap=1000., nfft=4000., detrend='constant',
#                 return_onesided=True, scaling='spectrum', axis=0)

freq = np.arange(0, w_fft.shape[0]) * fs / w_ave.shape[0]
fig = plt.figure()
ax = fig.add_subplot(121)
ax.plot(freq, np.abs(w_fft[:, 0]))
ax.plot(freq, np.abs(w_fft_2[:, 0]))
# ax.plot(f, px[:, 0]**0.5)
ax.plot(freq, 2*np.abs(np.fft.fft(s[:, 0, 1])[np.arange(0, w_fft.shape[0])])/s.shape[0])
plt.show()

fig = plt.figure()
ax = fig.add_subplot(121)
ax.plot(s[:, 0, 1])
ax.plot(w_ave[:, 0])
ax = fig.add_subplot(122)
ax.plot(s[:, 0, 1])
ax.plot(w_ave_2[:, 0])
plt.show()



import astropy.units as u
import pandas as pd
from pathlib import Path


def parse_eclipse_data(
        file_name: str | None = None,
        brick_keyword_left: str = 'LeftBricks',
        brick_keyword_right: str = 'RightBricks'
):
    """
    This function will read and return the data from Eclipse .xls files
    :param file_name: path to the file
    :param brick_keyword_left: string used to find if text file corresponds to left bricks
    :param brick_keyword_right: string used to find if text file corresponds to right bricks
    """
    file_name = Path(file_name)
    # get information about recording
    header = pd.read_csv(file_name, nrows=1, header=0, delimiter='\t', lineterminator='\n')
    header = header.loc[:, ~header.columns.str.contains('^Unnamed')]
    header = header.iloc[0]
    if file_name.name.find(brick_keyword_right) != -1:
        header['channel'] = 'right'
    elif file_name.name.find(brick_keyword_left) != -1:
        header['channel'] = 'left'
    else:
        header['channel'] = 'undefined_channel'
    n_bits = 16
    if isinstance(header['Volt/bit(nV)'], str):
        # check decimal point
        header['Volt/bit(nV)'] = float(header['Volt/bit(nV)'].replace(',', '.'))
    voltage_range = header['Volt/bit(nV)'] * 2 ** n_bits * u.nV
    print(voltage_range)
    # extract data in all buffers
    _data = pd.read_csv(file_name, skiprows=2, delimiter='\t', lineterminator='\n', index_col=0)
    _data_buffer_a = _data[_data.index.isin(['(A-buffer)'])]
    _data_buffer_b = _data[_data.index.isin(['(B-buffer)'])]
    # organize and scale data
    data = _data.to_numpy().T * header['Volt/bit(nV)'] * u.nV
    data = data[:, None, :].to(u.uV)

    data_buffer_a = _data_buffer_a.to_numpy().T * header['Volt/bit(nV)'] * u.nV
    data_buffer_a = data_buffer_a[:, None, :].to(u.uV)

    data_buffer_b = _data_buffer_b.to_numpy().T * header['Volt/bit(nV)'] * u.nV
    data_buffer_b = data_buffer_b[:, None, :].to(u.uV)
    return header, data, data_buffer_a, data_buffer_b



import numpy as np
from peegy.processing.tools import weightedAverage as averager, eeg_epoch_operators as eo
from peegy.processing.pipe.definitions import DataNode
from peegy.processing.tools.detection.definitions import TimeROI
import peegy.processing.tools.filters.eegFiltering as eegf
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.processing.tools.filters.eog_tools.tools import generate_eog_template, reconstruct_eog
from sklearn import linear_model
import matplotlib.pyplot as plt
import peegy.processing.tools.filters.spatial_filtering.spatial_filtering as sf
from peegy.processing.system.memory import enough_memory
import pyfftw
from scipy import signal
from sklearn.linear_model import LinearRegression
from peegy.tools.units.unit_tools import set_default_unit
import astropy.units as u
from tqdm import tqdm
from typing import List


__author__ = 'jaime undurraga'


def et_get_epoch_weights_and_rn_weighted_average(epochs: np.array = np.array([]),
                                                 block_size: int = 5,
                                                 samples_distance: int = 10,
                                                 demean_epochs: bool = True):
    """
    This function computes the weights and residual noise for weighted averaging.
    :param epochs: samples x channels x trials numpy array
    :param block_size: number of trials used to estimate noise and weights
    :param samples_distance: separation between samples to estimate noise and weights
    :param demean_epochs: if True, individual epochs will be demeaned
    :return: weights, residual noise, cumulative residual noise over trials,
    array with number of sources per noise source

    """
    # compute weights for averaging and residual noise estimation
    if demean_epochs:
        epochs = epochs - np.mean(epochs, 0)

    # compute weights for averaging and residual noise estimation
    block_size = np.minimum(epochs.shape[2], block_size)
    blocks = np.arange(0, epochs.shape[2] + 1, block_size).astype(int)
    _epochs = epochs[np.arange(0, epochs.shape[0], samples_distance), :, :]
    blocks[-1] = np.maximum(blocks[-1], _epochs.shape[2])
    unfolded_epochs = eo.et_unfold(np.transpose(_epochs, [0, 2, 1]), orthogonal=True)
    var = np.var(unfolded_epochs[:, blocks[0]: blocks[1]], ddof=1, axis=1)
    # var2 = np.expand_dims(np.mean(np.var(_epochs[:, :, blocks[0]: blocks[1]], ddof=1, axis=2), axis=0), axis=1)
    for i in range(1, len(blocks) - 1):
        var = np.vstack((var, np.var(unfolded_epochs[:, blocks[i]: blocks[i+1]], ddof=1, axis=1)))
    var = var[None, :]
    wb = 1.0 / np.mean(eo.et_fold(var.T, _epochs.shape[1]), axis=2)

    if np.all(wb == np.inf):
        wb[:] = 1
    nk = np.diff(blocks)
    w = np.ones((_epochs.shape[1], nk[0])) * np.expand_dims(wb[:, 0], 1)
    for i, n in enumerate(nk[1:]):
        w = np.concatenate((w, np.ones(n) * np.expand_dims(wb[:, i + 1], 1)), axis=1)
    rn = np.sqrt(1.0 / np.sum(w, axis=1))
    cumulative_rn = None
    for _i in range(w.shape[1]):
        if _i == 0:
            cumulative_rn = np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))
        else:
            cumulative_rn = np.vstack((cumulative_rn, np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))))

    n_total = np.sum(nk)
    return w, rn, cumulative_rn, n_total, wb, nk


def et_get_epoch_weights_and_rn_weighted_average_within_epoch(epochs: np.array = np.array([]),
                                                              demean_epochs: bool = True):
    """
    This function computes the weights and residual noise for weighted averaging by estimating the weights within each
    epoch. This is useful for long epochs (cortical responses) as artefacts, such as eye blinks, can be captured by the
    variance within epochs.
    :param epochs: samples x channels x trials numpy array
    :param demean_epochs: if True, individual epochs will be demeaned
    :return: weights, residual noise, cumulative residual noise over trials

    """
    if demean_epochs:
        epochs = epochs - np.mean(epochs, 0)
    # compute weights for averaging and residual noise estimation
    var = np.var(epochs, ddof=1, axis=0, keepdims=False)
    w = 1 / var
    rn = np.sqrt(1.0 / np.sum(w, axis=1))
    cumulative_rn = None
    for _i in range(w.shape[1]):
        if _i == 0:
            cumulative_rn = np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))
        else:
            cumulative_rn = np.vstack((cumulative_rn, np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))))
    n_total = epochs.shape[2]
    return w, rn, cumulative_rn, n_total


def et_get_epoch_weights_and_rn_standard_average(epochs: np.array = np.array([]),
                                                 block_size: int = 5,
                                                 samples_distance: int = 10,
                                                 demean_epochs: bool = True):
    """
    This function computes the weights and residual noise for standard averaging.
    :param epochs: samples x channels x trials numpy array
    :param block_size: number of trials used to estimate noise and weights
    :param samples_distance: separation between samples to estimate noise and weights
    :param demean_epochs: if True, individual epochs will be demeaned
    :return: weights, residual noise, cumulative residual noise over trials,
    array with number of sources per noise source
    """
    # compute weights for averaging and residual noise estimation
    if demean_epochs:
        epochs = epochs - np.mean(epochs, 0)

    # compute weights for averaging an residual noise estimation
    block_size = np.minimum(epochs.shape[2], block_size)
    blocks = np.arange(0, epochs.shape[2] + 1, block_size).astype(int)
    _epochs = epochs[np.arange(0, epochs.shape[0], samples_distance), :, :]
    blocks[-1] = np.maximum(blocks[-1], _epochs.shape[2])
    unfolded_epochs = eo.et_unfold(np.transpose(_epochs, [0, 2, 1]), orthogonal=True)
    var = np.var(unfolded_epochs[:, blocks[0]: blocks[1]], ddof=1, axis=1)[None, :]
    for i in range(1, len(blocks) - 1):
        var = np.vstack((var, np.var(unfolded_epochs[:, blocks[i]: blocks[i + 1]], ddof=1, axis=1)))
    nb_var = np.mean(eo.et_fold(var.T, _epochs.shape[1]), axis=2)
    wb = np.ones((_epochs.shape[1], var.shape[0]))
    nk = np.diff(blocks)
    w = np.ones((_epochs.shape[1], nk[0])) * np.expand_dims(wb[:, 0], 1) * u.dimensionless_unscaled
    for i, n in enumerate(nk[1:]):
        w = np.concatenate((w, np.ones(n) * np.expand_dims(wb[:, i + 1], 1)), axis=1)
    rn = np.sqrt(np.sum(nk * nb_var, axis=1) / (np.sum(nk) ** 2.0))
    cumulative_rn = None
    for _i in range(w.shape[1]):
        if _i == 0:
            cumulative_rn = np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))
        else:
            cumulative_rn = np.vstack((cumulative_rn, np.sqrt(1.0 / np.sum(w[:, 0:_i + 1], axis=1))))

    n_epochs = np.sum(nk)
    return w, rn, cumulative_rn, n_epochs, wb, nk


def et_mean(epochs: np.array = np.array([]),
            block_size: int = 10,
            samples_distance: int = 1,
            weighted: bool = True,
            normalize_weights: bool = False,
            demean_epochs: bool = True,
            weight_across_epochs: bool = True,
            n_jobs: int = 1):
    """
    Compute the mean of the input epochs
    :param epochs: samples x channels x trials numpy array
    :param block_size: number of trials used to estimate noise and weights
    :param samples_distance: separation between samples to estimate noise and weights
    :param weighted: boolean indicating if weighted average is used
    :param normalize_weights: if true weights will be normalized so that their sum across x is equal to 1
    :param demean_epochs: if True, individual epochs will be demeaned
    :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights are
    computed within epoch (1 / variance across time)
    :param n_jobs: number of CPUs to compute FFT
    :return: average data (standard or weighted), weights, residual noise, cumulative residual noise over trials,
    spectrum of final average scaled to time domain amplitudes, array with number of sources per noise source
    """
    epochs = set_default_unit(epochs, u.dimensionless_unscaled)
    if weighted:
        if weight_across_epochs:
            w, final_rn, cumulative_rn, n, wb, nb = et_get_epoch_weights_and_rn_weighted_average(
                epochs=epochs,
                block_size=block_size,
                samples_distance=samples_distance,
                demean_epochs=demean_epochs
            )
        else:
            wb, nb = None, None
            w, final_rn, cumulative_rn, n = et_get_epoch_weights_and_rn_weighted_average_within_epoch(
                epochs=epochs,
                demean_epochs=demean_epochs
            )
    else:
        w, final_rn, cumulative_rn, n, wb, nb = et_get_epoch_weights_and_rn_standard_average(
            epochs=epochs,
            block_size=block_size,
            samples_distance=samples_distance,
            demean_epochs=demean_epochs
        )
    # normalize weights to facilitate computations
    if normalize_weights:
        w = w / np.sum(w, axis=1)[:, None]
    w = np.tile(w, (epochs.shape[0], 1, 1))
    w_ave = eo.w_mean(epochs, w)
    fft = pyfftw.builders.rfft(w_ave,
                               overwrite_input=False,
                               planner_effort='FFTW_ESTIMATE',
                               axis=0,
                               threads=n_jobs)
    scaling_factor = 2 / epochs.shape[0]
    w_fft = fft() * scaling_factor * w_ave.unit
    snr = np.var(w_ave, ddof=1, axis=0) / final_rn ** 2 - 1
    return w_ave, w, final_rn, cumulative_rn, w_fft, n, wb, nb, snr


def et_snr_in_rois(data_node: DataNode = None,
                   roi_windows: np.array([TimeROI]) = None):
    data = data_node.data
    data_in_windows = get_in_window_data(data_node=data_node, roi_windows=roi_windows)
    rn = data_node.rn
    if rn is None:
        rn = [np.array([])]
    final_snr = np.zeros((data.shape[1], len(data_in_windows)))
    final_s_var = np.zeros((data.shape[1], len(data_in_windows)))
    f_values = np.zeros((data.shape[1], len(data_in_windows)))
    _snr = np.zeros((data.shape[1]))
    _fvalue = np.zeros((data.shape[1]))
    for _i, _data in enumerate(data_in_windows):
        _s_var = np.var(_data, ddof=1, axis=0).reshape(-1, )
        _noise_var = rn ** 2.0
        _idx_valid = _noise_var != 0
        _fvalue[_idx_valid] = _s_var[_idx_valid] / _noise_var[_idx_valid]
        _snr = np.maximum(_fvalue - 1.0, 0)
        _idx_invalid = _noise_var == 0
        _snr[_idx_invalid] = np.inf * u.dimensionless_unscaled
        _fvalue[_idx_invalid] = np.inf * u.dimensionless_unscaled
        final_snr[:, _i] = _snr
        final_s_var[:, _i] = _s_var
        f_values[:, _i] = _fvalue
    return final_snr, final_s_var, f_values


def get_in_window_data(data_node: DataNode = None,
                       roi_windows: np.array([TimeROI]) = np.array([TimeROI()])):
    roi_samples = [roi.get_roi_samples(data_node=data_node) for roi in roi_windows]
    output = []
    data = data_node.data
    for _i, _samples in enumerate(roi_samples):
        output.append(data[_samples, ...])
    return output


def et_snr_weighted_mean(averaged_epochs: np.array = np.array([]),
                         rn: np.array = np.array([]),
                         snr: np.array = np.array([]),
                         roi_windows: List[np.array] = None,
                         n_ch: int = None,
                         channel_idx: np.array = np.array([])):
    if not channel_idx.size:
        if n_ch is None:
            _v = np.arange(averaged_epochs.shape[1])
        else:
            _v = np.argsort(snr)[::-1][0:n_ch]
    else:
        _v = channel_idx
    _snr = snr
    if not np.any(snr):
        _snr = np.ones(snr.shape)
    w_ave = np.sum(averaged_epochs[:, _v] * _snr[_v].T, axis=1) / sum(_snr[_v])
    total_noise_var = rn[_v] ** 2.0
    w_total_noise_var = 1.0 / sum(1./total_noise_var)
    total_rn = w_total_noise_var ** 0.5
    total_snr, _signal_var = et_snr_in_rois(data=np.expand_dims(w_ave, 1), roi_windows=roi_windows, rn=total_rn)
    return np.expand_dims(w_ave, axis=1), total_rn, total_snr, _signal_var


def et_frequency_mean(epochs=np.array([]), fs=None,
                      samples_distance=10,
                      n_fft=None,
                      snr_time_window=np.array([]),
                      block_size=5,
                      test_frequencies=None,
                      scale_frequency_data=True,
                      weighted_average=True):
    j_ave = averager.JAverager()
    j_ave.splits = epochs.shape[1]
    j_ave.fs = fs
    j_ave.t_p_snr = np.arange(0, epochs.shape[0], epochs.shape[0] / samples_distance) / fs
    j_ave.analysis_window = snr_time_window
    j_ave.time_offset = 0.0
    j_ave.alpha_level = 0.05
    j_ave.min_block_size = block_size
    j_ave.plot_sweeps = False
    j_ave.frequencies_to_analyze = test_frequencies
    j_ave.n_fft = n_fft

    _n_sweep = 0
    _total_sweeps = epochs.shape[1] * epochs.shape[2]
    for i in range(epochs.shape[2]):
        for j in range(epochs.shape[1]):
            j_ave.add_sweep(epochs[:, j, i])
            _n_sweep += 1
            print(('averaged %i out of %i sweeps' % (_n_sweep, _total_sweeps)))
    if weighted_average:
        w_ave = j_ave.w_average
        w_fft = j_ave.w_fft
        rn = j_ave.w_rn
        snr = j_ave.w_snr
        s_var = j_ave.w_signal_variance
    else:
        w_ave = j_ave.s_average
        w_fft = j_ave.s_fft
        rn = j_ave.s_rn
        snr = j_ave.s_snr
        s_var = j_ave.s_signal_variance

    _h_test = j_ave.hotelling_t_square_test()
    if scale_frequency_data:
        n_samples = epochs.shape[0]
        n_fft = n_samples if n_fft is None else n_fft
        factor = 2 / np.minimum(n_samples, n_fft)
        w_fft = w_fft * factor
        for ch_tests in _h_test:
            for _s_test in ch_tests:
                _s_test.spectral_magnitude *= factor
                _s_test.rn *= factor
    return w_ave, rn, [snr], w_fft, _h_test, [s_var]


def et_frequency_wights(epochs=np.array([]), fs=None,
                        samples_distance=10,
                        n_fft=None,
                        snr_time_window=np.array([]),
                        block_size=5,
                        test_frequencies=None,
                        scale_frequency_data=True,
                        weighted_average=True):
    j_ave = averager.JAverager()
    j_ave.splits = epochs.shape[1]
    j_ave.fs = fs
    j_ave.t_p_snr = np.arange(0, epochs.shape[0], epochs.shape[0] / samples_distance) / fs
    j_ave.analysis_window = snr_time_window
    j_ave.time_offset = 0.0
    j_ave.alpha_level = 0.05
    j_ave.min_block_size = block_size
    j_ave.plot_sweeps = False
    j_ave.frequencies_to_analyze = test_frequencies
    j_ave.n_fft = n_fft

    _n_sweep = 0
    _total_sweeps = epochs.shape[1] * epochs.shape[2]
    for i in range(epochs.shape[2]):
        for j in range(epochs.shape[1]):
            j_ave.add_sweep(epochs[:, j, i])
            _n_sweep += 1
            print(('averaged %i out of %i sweeps' % (_n_sweep, _total_sweeps)))
    if weighted_average:
        w_ave = j_ave.w_average
        w_fft = j_ave.w_fft
        rn = j_ave.w_rn
        snr = j_ave.w_snr
        s_var = j_ave.w_signal_variance
    else:
        w_ave = j_ave.s_average
        w_fft = j_ave.s_fft
        rn = j_ave.s_rn
        snr = j_ave.s_snr
        s_var = j_ave.s_signal_variance

    _h_test = j_ave.hotelling_t_square_test()
    if scale_frequency_data:
        n_samples = epochs.shape[0]
        n_fft = n_samples if n_fft is None else n_fft
        factor = 2 / np.minimum(n_samples, n_fft)
        w_fft = w_fft * factor
        for ch_tests in _h_test:
            for _s_test in ch_tests:
                _s_test.spectral_magnitude *= factor
                _s_test.rn *= factor
    return w_ave, rn, [snr], w_fft, _h_test, [s_var]


def et_frequency_mean2(epochs=np.array([]),
                       fs: u.Quantity = None,
                       block_size: int = 10,
                       test_frequencies: np.array = None,
                       scale_frequency_data: bool = True,
                       weighted_average: bool = True,
                       delta_frequency: u.Quantity = 2 * u.Hz,
                       power_line_frequency: u.Quantity = 50 * u.Hz,
                       n_jobs: int = 1):
    epochs = set_default_unit(epochs, u.uV)
    fs = set_default_unit(fs, u.Hz)
    delta_frequency = set_default_unit(delta_frequency, u.Hz)
    test_frequencies = set_default_unit(test_frequencies, u.Hz)

    weights, pooled_rn, pooled_snr, by_freq_rn, by_freq_snr, freq_samples, exact_frequencies = \
        get_pooled_frequency_weights(
            epochs=epochs,
            fs=fs,
            block_size=block_size,
            frequencies=test_frequencies,
            weighted_average=weighted_average,
            delta_frequency=delta_frequency,
            power_line_frequency=power_line_frequency
        )

    w_ave = eo.w_mean(epochs=epochs,
                      weights=weights[0, :, :])
    fft = pyfftw.builders.rfft(w_ave, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                               threads=n_jobs)()
    scaling_factor = 1.0
    if scale_frequency_data:
        scaling_factor = 2 / epochs.shape[0]

    w_fft = fft * scaling_factor * w_ave.unit
    freq_samples = scaling_factor * freq_samples
    pooled_rn = scaling_factor * pooled_rn
    return w_ave, pooled_snr, pooled_rn, by_freq_snr, by_freq_snr, w_fft, weights, freq_samples, exact_frequencies


def et_subtract_correlated_ref(data: np.array = np.array([]),
                               idx_ref: np.array = np.array([]),
                               high_pass: float = 0.1,
                               low_pass: float = 20.0,
                               plot_results=False,
                               fs: float = 1.0,
                               figure_path: str = '',
                               figure_basename: str = ''):
    data = np.atleast_3d(data)
    origin_data = data.copy()
    fig = None
    # filter
    _b_h, _b_l = None, None
    if high_pass is not None:
        _b_h = eegf.bandpass_fir_win(high_pass=high_pass, low_pass=None, fs=fs)
    if low_pass is not None:
        _b_l = eegf.bandpass_fir_win(high_pass=None, low_pass=low_pass, fs=fs)
    for _trial in np.arange(data.shape[2]):
        reg = linear_model.LinearRegression()
        x = np.arange(0, data.shape[0])[:, None]
        reg.fit(x, data[:, :, _trial])
        data[:, :, _trial] = data[:, :, _trial] - reg.predict(x)
        temp_data = data[:, :, _trial]
        if high_pass is not None:
            temp_data = filt_data(data=data[:, :, _trial], b=_b_h)
        if low_pass is not None:
            temp_data = filt_data(data=temp_data, b=_b_l)
        ref = temp_data[:, idx_ref]
        for _idx in np.arange(ref.shape[1]):
            _current_ref = ref[:, _idx]
            transmission_index = temp_data.T.dot(_current_ref) / np.expand_dims(
                np.sum(np.square(_current_ref)), axis=0)
            data[:, :, _trial] -= _current_ref[:, None].dot(transmission_index[None, :])
            temp_data -= _current_ref[:, None].dot(transmission_index[None, :])
    figures = []
    if plot_results:
        time = np.arange(0, data.shape[0]) / fs
        fig = plt.figure()
        ax = fig.add_subplot(121)
        ax.plot(time, origin_data[:, 0, :].squeeze())
        ax.set_title('original')
        ax.legend(loc="upper right")
        ax = fig.add_subplot(122, sharey=ax)
        ax.plot(time, data[:, 0, :].squeeze())
        ax.set_title('eog removed')
        ax.legend(loc="upper right")
        fig.savefig(figure_path + figure_basename + '.png')
        figures.append(fig)
    return data, figures


def clean_peaks(peaks_amp: np.array = None,
                peaks_pos: np.array = None,
                minimum_width_samples: int = None):
    assert peaks_amp.size == peaks_pos.size
    _idx_diff, = np.where(np.append(0, np.diff(peaks_pos)) > minimum_width_samples)
    plt.plot(peaks_pos, peaks_amp)


def et_subtract_oeg_template(data: np.array = np.array([]),
                             fs: u.Quantity = None,
                             idx_ref: np.array = np.array([]),
                             high_pass: u.Quantity = 1 * u.Hz,
                             low_pass: u.Quantity = 20 * u.Hz,
                             template_width: u.Quantity = 1.4 * u.s,
                             crest_factor_threshold=10,
                             plot_results=False,
                             figure_path: str = '',
                             figure_basename: str = '',
                             minimum_interval_width: u.Quantity = 0.075 * u.s,
                             eog_peak_width: u.Quantity = 0.02 * u.s,
                             n_iterations: int = 10,
                             kernel_bandwidth: float = 0.15,
                             use_initial_template: bool = True):
    """
    This function removes EOG artefacts via a variation of the template subtraction method described in Valderrama
    et al., 2018.The algorithm performs several steps:
    1 - Blinks are detected in the EOG channel using a peak detection and generating a starting template.
    2 - The template is iteratively adjusted using  a math-filter.
    The location of the events is recomputed using the match filter and a new template is estimated over each iteration.
    3 - Once the template and the location of the EOG events are obtained, an EOG signal is generated by convolving the
    template with the location of the events (scaled to maximize the similarity between the template and the actual
    events.
    3 - The generated EOG signal is subtracted from the raw EEG signal, thus leading to a clean EEG signal.
    See Valderrama, J. T., de la Torre, A., & Van Dun, B. (2018). An automatic algorithm for blink-artifact suppression
    based on iterative template matching: Application to single channel recording of cortical auditory evoked
    potentials. Journal of Neural Engineering, 15(1), 016008. https://doi.org/10.1088/1741-2552/aa8d95

    The variations used in this implementation are
    1 - the initial template is generated from the data itself, without requiring any template.
    :param data: numpy array with the data to be cleaned (samples x channels) matrix
    :param fs: sampling rate
    :param idx_ref: index indicating the EOG reference channel
    :param high_pass: cutoff frequency for the high-pass filter
    :param low_pass: cutoff frequency for the low-pass filter
    :param template_width: width of the EOG template
    :param crest_factor_threshold: crest factor (in dB) used to detect and provide a first guess template from the data
    :param plot_results: if True, output plots will be generated
    :param figure_path: string indicating the path where to save the figures
    :param figure_basename: name of generated figure
    :param minimum_interval_width: minimum interval between eye blinks (only used for initial guess)
    :param eog_peak_width: empirical width of an EOG blink (only used for initial guess)
    :param n_iterations: number of iterations to improve the template estimation
    :param kernel_bandwidth: factor use to control the with of the Gaussian kernel in threshold detection
    :param use_initial_template: if true, a template eye blinking will be used as starting point to find events
    :return: clean data, figure
    """
    output_data = data.copy()
    # filter data
    if high_pass is not None or low_pass is not None:
        _b = eegf.bandpass_fir_win(high_pass=high_pass, low_pass=low_pass, fs=fs)
        data = filt_data(data=data, b=_b)

    original_ref = output_data[:, idx_ref]
    ref = data[:, idx_ref]

    # generate, reconstruct and remove EOG artefacts
    for _idx_eog_ch in range(ref.shape[1]):
        _current_eog = ref[:, [_idx_eog_ch]]
        template, eog_events, z = generate_eog_template(eog_data=_current_eog,
                                                        template_width=template_width,
                                                        fs=fs,
                                                        low_pass=low_pass,
                                                        eog_peak_width=eog_peak_width,
                                                        crest_factor_threshold=crest_factor_threshold,
                                                        minimum_interval_width=minimum_interval_width,
                                                        n_iterations=n_iterations,
                                                        kernel_bandwidth=kernel_bandwidth,
                                                        use_initial_template=use_initial_template)
        reconstructed = reconstruct_eog(data=data,
                                        template=template,
                                        blink_positions=eog_events)

        # remove from original data (untouched)
        output_data = output_data - reconstructed * data.unit
        # also remove from filtered data used to reconstruct, otherwise the reconstructed data will have EOG components
        # what were previously removed.
        data = data - reconstructed * data.unit
        figures = []
        if plot_results:
            time = np.arange(0, output_data.shape[0]) / fs
            time_template = (np.arange(0, template.shape[0]) - template.shape[0] / 2) / fs
            fig = plt.figure()
            ax = fig.add_subplot(311)
            ax.plot(time_template, template, label='fitted template')
            ax.legend(loc="upper right")

            ax = fig.add_subplot(312)
            ax.plot(time, _current_eog, label='reference eog')
            ax.plot(time, reconstructed[:, idx_ref[_idx_eog_ch]],
                    label='reconstructed eog')
            if eog_events.size:
                peak_pos = eog_events + np.argmax(template)
                ax.plot(time[peak_pos], reconstructed[peak_pos, idx_ref[_idx_eog_ch]],
                        'o')
            ax.legend(loc="upper right")

            ax = fig.add_subplot(313)
            ax.plot(time, original_ref[:, _idx_eog_ch], label='reference eog')
            ax.plot(time, output_data[:, idx_ref[_idx_eog_ch]], label='eog removed')
            ax.legend(loc="upper right")
            fig.savefig(figure_path + figure_basename + 'eog_ref_{:}.png'.format(_idx_eog_ch))
            figures.append(fig)
    return output_data, figures, template, z, eog_events


def et_get_spatial_filtering(epochs: np.array = np.array([]),
                             fs: u.Quantity = None,
                             sf_join_frequencies: np.array = None,
                             weight_data: bool = True,
                             weighted_frequency_domain: bool = False,
                             weight_across_epochs: bool = True,
                             weights: np.array = None,
                             demean_data: bool = True,
                             keep0: int = None,
                             keep1: float = 1e-9,
                             perc0: float = 1,
                             perc1: float = None,
                             block_size: int = 10,
                             n_tracked_points: int = None,
                             delta_frequency: u.Quantity = 5 * u.Hz,
                             n_jobs: int = 1,
                             ):
    """

    :param epochs: m x n x l (samples, channels, epochs)
    :param fs: sampling rate
    :param sf_join_frequencies: numpy array with frequencies to bias the filter
    :param weight_data: if True, covariance and mean will be estimated using weights
    :param weighted_frequency_domain: boll indicating if the weghts are compute in the time or frequency domain
    :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
    are computed within epoch (1 / variance across time)
    :param weights: array of weights to use when weighted_average is true. If none and weighted_average is true,
    these will be estimated.
    :param demean_data: if True, data will be demeaned prior weights estimation.
    :param keep0: integer controlling  whitening of unbiased components in DSS. This integer value represent the number
    of components to keep.
    :param keep1: float controlling  whitening of unbiased components in DSS. This value will remove components below
    keep1 which is relative to the maximum eigen value.
    :param perc0: float (between 0 and 1) controlling whitening of unbiased components in DSS.
    This value will preserve components that explain the percentage of variance.
    :param perc1: float (between 0 and 1) controlling the number of biased components kept in DSS.
    This value will preserve the components of the biased PCA analysis that explain the percentage of variance.
    :param block_size: integer indicating the number of trials that would be use to estimate the weights
    :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
    :param delta_frequency: frequency size around each sf_join_frequency to estimate noise
    :param n_jobs: number of CPUs to compute FFT
    :return: z, pwr0, pwr1, cov_1, weights: components, unbiased power, biased power, covariance rotation,
    estimated weights
    """
    print('computing spatial filter')

    # bias function uses weighted mean or standard mean
    if weight_data and weights is None:
        if weighted_frequency_domain:
            weights, *_ = get_pooled_frequency_weights(
                epochs=epochs,
                fs=fs,
                frequencies=sf_join_frequencies,
                weighted_average=weight_data,
                block_size=block_size,
                delta_frequency=delta_frequency
            )
        else:
            if n_tracked_points is None:
                n_tracked_points = epochs.shape[0]
            samples_distance = int(max(epochs.shape[0] // n_tracked_points, 1))
            _, weights, *_ = et_mean(epochs,
                                     weighted=weight_data,
                                     weight_across_epochs=weight_across_epochs,
                                     block_size=block_size,
                                     samples_distance=samples_distance)

    if weights is None:
        weights = np.ones(epochs.shape) * u.dimensionless_unscaled
    if demean_data:
        epochs = eo.et_demean(epochs, w=weights)

    average = eo.w_mean(epochs, weights=weights)

    if sf_join_frequencies is not None:
        n_frequencies = sf_join_frequencies / fs
        c0, t0 = eo.et_freq_weighted_cov(epochs,
                                         normalized_frequencies=n_frequencies,
                                         n_jobs=n_jobs)
        c1, t1 = eo.et_freq_weighted_cov(average,
                                         normalized_frequencies=n_frequencies,
                                         n_jobs=n_jobs)
        c0 = c0 / t0
        c1 = c1 / t1
        todss, pwr0, pwr1, n_0, n_1 = sf.nt_dss0(c0, c1, keep0=keep0, keep1=keep1, perc0=perc0, perc1=perc1)
        z = eo.et_mmat(epochs, todss)
        cov_1, tw_cov = eo.et_freq_shifted_xcovariance(z, epochs,
                                                       normalized_frequencies=n_frequencies,
                                                       wy=weights.value,
                                                       n_jobs=n_jobs)
        cov_1 = cov_1 / tw_cov
    else:
        # clean data using dss with weighted average as bias. Data covariance weighted with same weights
        c0, tw0 = eo.et_weighted_covariance(epochs)
        c1, tw1 = eo.et_weighted_covariance(average)
        c0 = c0 / tw0
        c1 = c1 / tw1
        todss, pwr0, pwr1, n_0, n_1 = sf.nt_dss0(c0, c1, keep0=keep0, keep1=keep1, perc0=perc0, perc1=perc1)
        z = eo.et_mmat(epochs, todss)
        cov_1, tw_cov = eo.et_time_shifted_xcovariance(z,
                                                       epochs,
                                                       wy=weights.value
                                                       )
        cov_1 = cov_1 / tw_cov

    return z, pwr0, pwr1, cov_1, n_0, n_1, weights


def et_apply_spatial_filtering(z=np.array([]),
                               pwr0=np.array([]),
                               pwr1=np.array([]),
                               cov_1=np.array([]),
                               sf_components=np.array([]),
                               sf_thr=0.8):
    print('applying spatial filter')
    max_num_components = z.shape[1]
    if sf_components is None or not sf_components.size:
        # threshold is set to account for st_thr variance of the evoked power
        pos = np.maximum(np.where(np.cumsum(pwr1) / np.sum(pwr1) >= sf_thr)[0][0], 1)
        n_components = np.arange(pos)
    else:
        n_components = sf_components
    n_components = n_components[n_components < max_num_components]
    cleaned_data = eo.et_mmat(z[:, n_components, :], cov_1[n_components, :])
    _total_explained_var = (np.sum(pwr1[n_components]) / np.sum(pwr0[n_components])) / np.sum(pwr1 / pwr0)
    _evoked_explained_var = np.sum(pwr1[n_components]) / np.sum(pwr1)
    print('DSS keeping {:} out of {:} components accounting for {:}% of the total variance and {:}% of '
          'evoked variance'.format(n_components.size,
                                   pwr0.size,
                                   np.round(_total_explained_var * 100, decimals=2),
                                   np.round(_evoked_explained_var * 100, decimals=2)
                                   ))
    return cleaned_data, n_components


def estimate_rn_from_trace(data=np.array([]), rn_ini_sample=0, rn_end_sample=-1, plot_estimation=False):
    x = np.expand_dims(np.arange(0, data[rn_ini_sample:rn_end_sample, :].shape[0]), axis=1)
    _subset = data[rn_ini_sample:rn_end_sample, :]
    regression = linear_model.LinearRegression()
    regression.fit(x, _subset)
    rn = np.std(_subset - regression.predict(x), axis=0)
    if plot_estimation:
        plt.plot(x, _subset)
        plt.plot(x, regression.predict(x))
        plt.title('rn regression, rn:' + str(rn))
        plt.show()

    return rn


def estimate_srn_from_trace(data=np.array([]), snr_ini_sample=0, snr_end_sample=-1, **kwargs):
    rn = estimate_rn_from_trace(data=data, **kwargs)
    _subset = data[snr_ini_sample:snr_end_sample, :]
    snr = np.var(_subset, ddof=1, axis=0) / rn ** 2.0 - 1
    return snr, rn


def eye_artifact_subtraction(epochs: np.array = np.array([]),
                             oeg_epochs: np.array = np.array([])):
    # subtract eye artifacts
    print('removing oeg artifacts')
    clean_data = et_subtract_correlated_ref(epochs=epochs, ref=oeg_epochs)
    return clean_data


def et_average_frequency_transformation(epochs: np.array = np.array([]),
                                        fs=0.0,
                                        ave_mode='magnitude',
                                        n_jobs: int = 1):

    fft = pyfftw.builders.rfft(epochs, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                               threads=n_jobs)
    _fft = fft()

    if ave_mode == 'magnitude':
        _fft = np.abs(_fft)
    ave = np.abs(np.mean(_fft, axis=2))
    ave *= 2/epochs.shape[0]
    freqs = np.arange(0, _fft.shape[0]) * fs / epochs.shape[0]
    return ave, freqs


def bootstrap(data: np.array = np.array([]), statistic=np.mean, num_samples=1000, alpha=0.05):
    """
    This function bootstrap incoming data using passed statistic function
    :param data: numpy array
    :param statistic: function that will be bootstrapped
    :param num_samples: number of estimations that will be used
    :param alpha: probabilistic level to compute confidence intervals
    :param axis: axis where statistic will be applied
    :return: bootstrap estimate of 100.0*(1-alpha) CI for statistic.
    """
    n = data.shape[2]
    idx = np.random.randint(0, n, (int(num_samples), n))
    unfolded_epochs = eo.et_unfold(np.transpose(data, [0, 2, 1]), orthogonal=True)
    new_data = np.zeros((unfolded_epochs.shape[0], num_samples))
    for _set in range(num_samples):
        samples = unfolded_epochs[:, idx[_set]]
        new_data[:, _set] = statistic(samples, axis=1)

    stat = np.sort(new_data, axis=1)
    mean = np.mean(stat, axis=1)
    ci_low, ci_hi = (stat[:, int((alpha / 2.0) * num_samples)], stat[:, int((1 - alpha / 2.0) * num_samples)])
    out_mean = np.reshape(mean, (data.shape[0], data.shape[1]), order='F')
    out_ci_low = np.reshape(ci_low, (data.shape[0], data.shape[1]), order='F')
    out_ci_hi = np.reshape(ci_hi, (data.shape[0], data.shape[1]), order='F')
    return out_mean, out_ci_low, out_ci_hi


def et_impulse_response_artifact_subtraction(
        data: np.array = None,
        stimulus_waveform: np.array = None,
        ir_length: int = 100,
        ir_max_lag: int = 0,
        regularization_factor: float = 1,
        plot_results: bool = False,
        n_jobs: int = 1
) -> np.array:
    """
    This function will remove an artifact which has the shape of the input stimulus_waveform.
    Here, artifact is estimated from the impulse response. The latter is estimated by averaging individual
    impulse responses across epochs.
    :param data: data samples x channels x trials
    :param stimulus_waveform: single column numpy array with the target waveform artifact
    :param ir_length: estimated impulse response length in samples
    :param ir_max_lag: maximum lag or sydtem delay (from zero). Impulse response window will be centered around the max
    within the max_lag.
    :param regularization_factor: This value is used to regularize the deconvolution based on Tikhonov regularization
    allow to estimate the transfer function when the input signal is sparse in frequency components
    :param plot_results: if True, figures with results will be shown
    :param n_jobs: number of CPUs to compute FFT
    :return: data without artifacts
    """
    original_data_shape = data.shape
    # compute impulse response; data padded to next power of 2 to speed process
    n_fft = next_power_two(data.shape[0])

    data = np.atleast_3d(data)
    _stim_template = stimulus_waveform[0: data.shape[0]]
    data = np.pad(data, ((0, n_fft - data.shape[0]), (0, 0), (0, 0)), constant_values=(0, 0), mode='constant')
    _stim_template = np.pad(_stim_template, ((0, n_fft - _stim_template.shape[0]), (0, 0)),
                            constant_values=(0, 0), mode='constant')
    _fft = pyfftw.builders.rfft(_stim_template,
                                overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                threads=n_jobs)
    x_fft = _fft()
    _fft = pyfftw.builders.rfft(data,
                                overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                threads=n_jobs)
    y_fft = _fft()

    # estimate impulse response for each epoch via Tikhonov regularization
    x_fft = np.atleast_3d(x_fft)
    h_fft = y_fft * np.conjugate(x_fft) / (x_fft * np.conjugate(x_fft) + regularization_factor)

    _ifft = pyfftw.builders.irfft(h_fft,
                                  n=data.shape[0],
                                  overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                  threads=n_jobs)
    h = _ifft()
    h = np.mean(h, axis=2, keepdims=True)
    w = np.zeros(h.shape)
    ws = np.repeat(signal.windows.hanning(ir_length).reshape(-1, 1), h.shape[1], axis=1)
    w[0: ws.shape[0], :, :] = np.atleast_3d(ws)
    _max_ir = np.argmax(h[0: ir_max_lag + 1])
    w = np.roll(w, -ir_length // 2 + _max_ir, axis=0)
    h_ave_w = h * w
    # estimate artifact by convolution
    _fft = pyfftw.builders.rfft(h_ave_w,
                                overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                threads=n_jobs)
    h_ave_w_fft = _fft()
    # convolve
    fft_artifact = h_ave_w_fft * x_fft
    _ifft = pyfftw.builders.irfft(fft_artifact,
                                  n=data.shape[0],
                                  overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                  threads=n_jobs)
    _recovered_artifact = _ifft() * data.unit
    _clean_data = data - _recovered_artifact
    if plot_results:
        plt.figure()
        plt.plot(h.squeeze(), label='full IR')
        plt.plot(h_ave_w.squeeze(), label='windowed IR')
        plt.legend()
        plt.show()
    clean_data = np.resize(_clean_data, original_data_shape)
    recovered_artifact = np.resize(_recovered_artifact, original_data_shape)
    return clean_data, recovered_artifact


def next_power_two(n):
    # Returns next power of two following 'number'
    return int(2 ** np.ceil(np.log2(n)))


def et_regression_subtraction(
        data: np.array = None,
        stimulus_waveform: np.array = None,
        max_lag: int = 0,
        max_length: int = None,
        plot_results: bool = False
) -> np.array:
    """
    This function will remove an artifact which has the shape of the input stimulus_waveform.
    Here, artifact is estimated from the impulse response. The latter is estimated by averaging individual
    impulse responses across epochs.
    :param data: data samples x channels x trials
    :param stimulus_waveform: single column numpy array with the target waveform artifact
    :param max_lag: max expected delay between artifact and stimulus waveform. The maximum correlation will be searched
     within this  0 to max_lag
    :param max_length: determines the maxlegnth to search for correlation between stimulus artifact and data. This is
    useful if the desire response has a delay relative to the stimulus artifact. You could limit the correlation to look
    only on the region containing only the artifact,
    :param plot_results: if True, figures with results will be shown
    :return: data without artifacts
    """
    clean_data = np.zeros(data.shape)
    subset_idx = np.arange(0, data.shape[0])
    if max_length is not None:
        subset_idx = np.arange(0, np.minimum(max_length, data.shape[0]))
    sub_stimulus_waveform = stimulus_waveform[subset_idx]
    sub_data_set = data[subset_idx, :]
    _optimal_lag = 0
    # first we detect the lag to which regression fit is maximal across channels
    if max_lag > 0:
        score = np.zeros((max_lag, data.shape[1]))
        for _idx in np.arange(data.shape[1]):
            for _i_lag in range(max_lag):
                regression = LinearRegression()
                _delayed_sub_stimulus_waveform = np.pad(sub_stimulus_waveform,
                                                        ((_i_lag, 0),
                                                         (0, 0)), 'constant', constant_values=(0, 0))
                _delayed_sub_stimulus_waveform.resize(sub_stimulus_waveform.shape, refcheck=False)
                regression.fit(_delayed_sub_stimulus_waveform, sub_data_set[:, _idx])
                score[_i_lag, _idx] = regression.score(_delayed_sub_stimulus_waveform, sub_data_set[:, _idx])
        # we define the optimal lag as the average one across all channel
        _optimal_lag = np.round(np.mean(np.argmax(score, axis=0))).astype(int)

    # now we use optimal lag to subtract artifact
    for _idx in tqdm(np.arange(data.shape[1]), desc='Artefact regression'):
        regression = LinearRegression()
        _delayed_sub_stimulus_waveform = np.pad(sub_stimulus_waveform,
                                                ((_optimal_lag, 0),
                                                 (0, 0)), 'constant', constant_values=(0, 0))
        _delayed_sub_stimulus_waveform.resize(sub_stimulus_waveform.shape, refcheck=False)
        regression.fit(_delayed_sub_stimulus_waveform, sub_data_set[:, _idx])
        delayed_stimulus_waveform = np.pad(stimulus_waveform,
                                           ((_optimal_lag, 0),
                                            (0, 0)), 'constant', constant_values=(0, 0))
        delayed_stimulus_waveform.resize(stimulus_waveform.shape, refcheck=False)
        recovered_artifact = regression.predict(delayed_stimulus_waveform)
        clean_data[:, _idx] = data[:, _idx] - recovered_artifact

    if plot_results:
        plt.plot(data[:, -1], label="original data")
        plt.plot(recovered_artifact[:, -1], label="recovered artifact")
        plt.plot(clean_data[:, -1], label="clean data")
        plt.legend()
        plt.show()
    return clean_data, recovered_artifact


def et_xcorr_subtraction(
        data: np.array = None,
        stimulus_waveform: np.array = None,
        max_lag: int = 0,
        max_length: int = None,
        plot_results: bool = False,
) -> np.array:
    """
    This function will remove an artifact which has the shape of the input stimulus_waveform.
    Here, artifact is estimated from the impulse response. The latter is estimated by averaging individual
    impulse responses across epochs.
    :param data: data samples x channels x trials
    :param stimulus_waveform: single column numpy array with the target waveform artifact
    :param max_lag: max expected delay between artifact and stimulus waveform. The maximum correlation will be searched
     within this  0 to max_lag as we assume only a positive delay
    :param max_length: determines the maxlegnth to search for correlation between stimulus artifact and data. This is
    useful if the desire response has a delay relative to the stimulus artifact. You could limit the correlation to look
    only on the region containing only the artifact,
    :param plot_results: if True, figures with results will be shown
    :return: data without artifacts
    """
    original_shape = data.shape
    data = np.atleast_3d(data)
    clean_data = np.zeros(data.shape)
    subset_idx = np.arange(0, data.shape[0])
    if max_length is not None:
        subset_idx = np.arange(0, np.minimum(max_length, data.shape[0]))
    sub_stimulus_waveform = stimulus_waveform[subset_idx]
    sub_data_set = data[subset_idx, ::]
    _optimal_lag = 0
    if max_lag > 0:
        score = np.zeros((max_lag, data.shape[1]))
        for _i_lag in range(max_lag):
            _delayed_sub_stimulus_waveform = np.pad(sub_stimulus_waveform,
                                                    ((_i_lag, 0),
                                                     (0, 0)), 'constant', constant_values=(0, 0))
            _delayed_sub_stimulus_waveform.resize(sub_stimulus_waveform.shape, refcheck=False)
            ave_data = np.mean(sub_data_set, 2)
            transmission_index = ave_data.T.dot(_delayed_sub_stimulus_waveform) / np.expand_dims(
                np.sum(np.square(_delayed_sub_stimulus_waveform)), axis=0)
            recovered_artifact = np.tile(np.atleast_3d(stimulus_waveform * transmission_index.T), (1, 1, data.shape[2]))
            clean_data = data - recovered_artifact
            score[_i_lag, :] = np.mean(np.std(np.mean(data - recovered_artifact, axis=2), axis=0))
        # we define the optimal lag as the average one across all channel
        _optimal_lag = np.round(np.mean(np.argmax(score, axis=0))).astype(int)

    # remove artifact at optimal lag
    _delayed_sub_stimulus_waveform = np.pad(sub_stimulus_waveform,
                                            ((_optimal_lag, 0),
                                             (0, 0)), 'constant', constant_values=(0, 0))
    _delayed_sub_stimulus_waveform.resize(sub_stimulus_waveform.shape, refcheck=False)
    ave_data = np.mean(sub_data_set, 2)
    transmission_index = ave_data.T.dot(_delayed_sub_stimulus_waveform) / np.expand_dims(
        np.sum(np.square(_delayed_sub_stimulus_waveform)), axis=0)

    delayed_stimulus_waveform = np.pad(stimulus_waveform,
                                       ((_optimal_lag, 0),
                                        (0, 0)), 'constant', constant_values=(0, 0))
    delayed_stimulus_waveform.resize(stimulus_waveform.shape, refcheck=False)

    recovered_artifact = np.tile(np.atleast_3d(delayed_stimulus_waveform * transmission_index.T),
                                 (1, 1, data.shape[2]))
    clean_data = data - recovered_artifact

    if plot_results:
        plt.plot(data[:, -1], label="original data")
        plt.plot(recovered_artifact[:, -1], label="recovered artifact")
        plt.plot(clean_data[:, -1], label="clean data")
        plt.legend()
        plt.show()
    clean_data = clean_data.reshape(original_shape)
    return clean_data, recovered_artifact


def et_ica_epochs(
        data: np.array = np.array([]),
        tol: float = 1e-4,
        iterations: int = 200):
    """
    This function computes and sorts the components of an input matrix using ICA. The components are sorted by their
    power.
    :param data: np.array where components are obtained for each column
    :param tol: float indicating the tolerance for the difference between previous and updated weights to stop the
    iteration
    :param iterations: maximum number of iterations to estimate the components
    :return: components, unmixing matrix (used to obtain the components), mixing_matrix (used to project the components
    back to the original space, power of components, whitening matrix
    """

    no_chans = data.shape[1]
    whitened_ave, whitening_m = et_ica_whitening(data)

    # Implement ICA
    # Guess random initial value for W (inverse mixing matrix) and update until it is orthogonal
    # (ie. multiplied by transverse = 1)
    # Initialise array for w
    w = np.zeros((no_chans, no_chans))

    # Iterate through W matrix and update it - ie. calculate new value for w
    # Based on formula from https://towardsdatascience.com/independent-component-analysis-ica-in-python-a0ef0db0955e
    for i in range(no_chans):
        W = np.random.rand(no_chans)

        j = 0
        tol_lim = 1

        while (j < iterations) and (tol_lim > tol):
            u = np.dot(W, whitened_ave)
            g = np.tanh(u)
            g_dash = 1 - np.square(np.tanh(u))
            W_update = (whitened_ave * g).mean(axis=1) - (g_dash.mean() * W.squeeze())
            # Decorrelating w
            W_update = W_update - np.dot(np.dot(W_update, w[:i].T), w[:i])
            W_update = W_update / np.sqrt((W_update ** 2).sum())
            # Create a limit condition using the current value of w
            tol_lim = np.abs(np.abs((W_update * W).sum()) - 1)
            # Update original w with the new values
            W = W_update
            # Iterator counter
            j += 1
        w[i, :] = W.T

    # Unmix Signals
    components = np.dot(w, whitened_ave)
    components = components.T
    mixing = np.linalg.pinv(np.dot(w, whitening_m))

    # Sorting by power
    # Calculate and sort a power array (large to small)
    pwr = np.max(np.abs(mixing), axis=0)
    idx_pwr = np.argsort(-pwr)
    s_pwr = pwr[idx_pwr]

    # Sort mixing matrix and components by power
    unmixing = w.T
    s_unmixing = unmixing[:, idx_pwr].T
    s_mixing = mixing[:, idx_pwr]
    s_components = components[:, idx_pwr]

    # Projection back into cortical space
    # proj = (np.dot(mixing, components.T)).T

    return s_components, s_unmixing, s_mixing, s_pwr, whitening_m


def et_ica_whitening(data: np.array = None):
    # Preprocessing
    data = data.T

    # Centering - subtracting the mean from the observed data
    mean = np.mean(data, axis=1, keepdims=True)
    centered_data = data - mean

    # Whitening x - removing correlations between components
    # Calculate covariance matrix
    covar = (centered_data.dot(centered_data.T)) / (np.shape(centered_data)[1] - 1)

    # Use single value decomposition to describe the matrix using its constituents
    # Breaking it down into left singular vectors (U) or eigenvectors, the singular values or eigenvalues (K)
    # and right singular vectors (V)
    u, k, v = np.linalg.svd(covar)

    # Create a diagonal matrix of the eigenvalues (inverse square)
    d = np.diag(1 / np.sqrt(k))

    # Calculate the matrix which will whiten the weight averaged epochs
    whitening_m = np.dot(u, np.dot(d, u.T))

    # Whiten EEG averaged data using above whitening matrix
    whitened_ave = np.dot(whitening_m, data)

    return whitened_ave, whitening_m


def goertzel(data, fs, frequency):
    """
    Implementation of the Goertzel algorithm, to calculate individual frequency Fourier values.
    :param data: (time x channels) numpy arra
    :param fs: float indicating the sampling rate
    :param frequency: frequency to be estimated in Hz
    :return: np.array with Fourier frequency term for each channel (1 x channels)
    Frequency amplitude will match that of numpy.fft.fft, exact frequency bin tested
    """
    n = data.shape[0]
    freqs = np.arange(0, n) * fs / n
    k = np.argmin(np.abs(freqs - frequency))
    cr = np.cos(2.0 * np.pi * k / n)
    ci = np.sin(2.0 * np.pi * k / n)
    coeff = 2.0 * cr
    sprev = 0.0
    sprev2 = 0.0
    for _i in tqdm(np.arange(n).astype(int), desc='DFT estimation {:}'.format(frequency)):
        s = data[[_i], :] + coeff * sprev - sprev2
        sprev2 = sprev
        sprev = s
    real = sprev * cr - sprev2
    imag = sprev * ci
    exact_frequency = freqs[k]
    return real + 1j * imag, exact_frequency


def discrete_dft(data: np.array = None,
                 fs: float = None,
                 frequencies: float = None,
                 method: str = 'auto',
                 n_jobs: int = 1):
    """
    Gets individual frequency Fourier values using either fast fft or Goertzel depending on available memory.
    :param data: (time x channels) numpy array
    :param fs: float indicating the sampling rate
    :param frequencies: frequencies to be estimated in Hz
    :param method: string specifying how to compute DFT
    :param n_jobs: number of CPUs to comput FFT
    :return: np.array with Fourier frequency term for each channel (1 x channels)
    """
    data = set_default_unit(data, u.uV)
    frequencies = set_default_unit(frequencies, u.Hz)
    n = data.shape[0]
    freqs = np.arange(0, n) * fs / n
    k = np.array([np.argmin(np.abs(freqs - _f)) for _f in frequencies])
    dft = np.zeros((k.size, data.shape[1], data.shape[2])) * data.unit
    exact_frequencies = np.zeros(k.size) * frequencies.unit
    _memory = enough_memory(data)
    if method == 'auto':
        _method = 'fft' if _memory else 'goertzel'

    if _method == 'fft':
        fft = pyfftw.builders.rfft(data,
                                   overwrite_input=False,
                                   planner_effort='FFTW_ESTIMATE',
                                   axis=0,
                                   threads=n_jobs)
        _fft = fft()
        dft = _fft[k, :] * data.unit
        exact_frequencies = freqs[k]
    elif _method == 'goertzel':
        for _i, _f in enumerate(frequencies):
            _dft, _exact_frequency = goertzel(data, fs, _f)
            dft[_i] = _dft
            exact_frequencies[_i] = _exact_frequency
    else:
        raise 'no implemented method'

    return dft, exact_frequencies


def get_discrete_frequencies_weights(
        epochs: np.array = None,
        fs: float = None,
        frequencies: float = None,
        block_size: int = 5,
        weighted_average=True) -> np.array:
    """
    Compute spectral amplitude for a given frequencies
    component.
    :param epochs: time x channels x trials numpy array in the time domain
    :param fs: float indicating the sampling rate in Hz
    :param frequencies: array with the frequencies of interest
    :param block_size: integer indicating the number of trials that would be use to estimate the weights
    :param weighted_average: if false, frequency bins will not be weighted to compute mean and std
    """
    epochs = set_default_unit(epochs, u.uV)
    fs = set_default_unit(fs, u.Hz)
    frequencies = set_default_unit(frequencies, u.Hz)

    y_fft, exact_frequencies = discrete_dft(epochs, fs, frequencies)
    # y_noise, exact_frequencies = discrete_dft(epochs - np.mean(epochs, axis=2, keepdims=True), fs, frequencies)
    block_size = np.minimum(epochs.shape[2], block_size)
    blocks = np.arange(0, epochs.shape[2] + 1, block_size).astype(int)
    blocks[-1] = np.maximum(blocks[-1], epochs.shape[2])
    n_blocks = np.maximum(len(blocks) - 1, 1)
    r = np.zeros((y_fft.shape[0], epochs.shape[1], n_blocks))
    norm_r = np.zeros((y_fft.shape[0], epochs.shape[1], n_blocks))
    total_var = np.zeros((y_fft.shape[0], epochs.shape[1], n_blocks)) * epochs.unit ** 2.0
    # r = np.zeros((y_noise.shape[0], epochs.shape[1], n_blocks))
    # norm_r = np.zeros((y_noise.shape[0], epochs.shape[1], n_blocks))
    # total_var = np.zeros((y_noise.shape[0], epochs.shape[1], n_blocks)) * epochs.unit ** 2.0
    wb = np.ones(total_var.shape) * total_var.unit
    ini_block, end_block = blocks[:-1], np.cumsum(np.diff(blocks))[::1]
    for i, (_ini_trial, _end_trial) in enumerate(zip(ini_block, end_block)):
        mag = np.abs(y_fft[:, :, _ini_trial: _end_trial])
        ang = np.angle(y_fft[:, :, _ini_trial: _end_trial]).value
        # mag = np.abs(y_noise[:, :, _ini_trial: _end_trial])
        # ang = np.angle(y_noise[:, :, _ini_trial: _end_trial]).value
        if weighted_average:
            angular_weights = mag ** 2.0
            angular_weights = angular_weights / np.sum(angular_weights, axis=2, keepdims=True)
        else:
            angular_weights = 1.0

        norm_r[:, :, i] = np.abs(np.sum((angular_weights * np.exp(1j * ang)), axis=2))
        mean_x = np.sum((mag * angular_weights * np.cos(ang)), axis=2, keepdims=True)
        mean_y = np.sum((mag * angular_weights * np.sin(ang)), axis=2, keepdims=True)
        r[:, :, i] = np.sqrt(mean_x ** 2.0 + mean_y ** 2)[:, :, 0]
        var_x = np.sum(angular_weights * (mag * np.cos(ang) - mean_x) ** 2.0, axis=2)
        var_y = np.sum(angular_weights * (mag * np.sin(ang) - mean_y) ** 2.0, axis=2)
        # we assumed that variance on x and y axis are independent from each other
        # this encapsulates the variance: sigma_r^2 <= variance <= sigma_r^2 + 4 * sum(w_angular*r_i*mean(r_i))
        total_var[:, :, i] = var_x + var_y
        # purely radial
        # mean_r = np.sum(mag * angular_weights, axis=2, keepdims=True)
        # total_var[:, :, i] = np.sum(angular_weights * (mag - mean_r) ** 2.0, axis=2)

        # fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
        # ax.plot(ang.T.squeeze(), mag.T.squeeze(), 'o')
        # ax.plot(np.angle(mean_x + 1j*mean_y).squeeze(), r[:, :, i].squeeze(), 'o', markersize=8, color='k')

    if weighted_average:
        wb = 1 / total_var
    if np.all(wb == np.inf):
        wb[:] = 1
    w = np.zeros(y_fft.shape) * wb.unit
    # w = np.zeros(y_noise.shape) * wb.unit
    for i, (_ini_trial, _end_trial) in enumerate(zip(ini_block, end_block)):
        w[:, :, _ini_trial: _end_trial] = wb[:, :, [i]]
    if weighted_average:
        rn = np.sqrt(1.0 / np.sum(block_size / total_var, axis=2))
    else:
        rn = np.sqrt(np.sum(block_size * total_var, axis=2) / ((block_size * n_blocks) ** 2.0))
    spectral_magnitude = np.abs(eo.w_mean(y_fft, weights=w))
    sig_plus_noise_var = spectral_magnitude ** 2.0
    snr = sig_plus_noise_var / (rn ** 2) - 1
    return w, spectral_magnitude, sig_plus_noise_var, rn, snr, exact_frequencies, y_fft


def get_pooled_frequency_weights(
        epochs: np.array = None,
        fs: float = None,
        frequencies: float = None,
        block_size: int = 5,
        weighted_average=True,
        delta_frequency: u.Quantity = 5 * u.Hz,
        power_line_frequency: u.Quantity = 50 * u.Hz) -> np.array:
    """
    Compute spectral amplitude and pooled weights for set of frequencies.
    :param epochs: time x channels x trials numpy array in the time domain
    :param fs: float indicating the sampling rate in Hz
    :param frequencies: numpy array indicating the frequency of interest
    :param block_size: integer indicating the number of trials that would be use to estimate the weights
    :param weighted_average: if false, frequency bins will not be weighted to compute mean and std
    :param delta_frequency: frequency size around each frequency used to estimate the noise
    :param power_line_frequency: frequency of local power line frequency. This will be used to prevent using
        this frequency or its multiples when performing frequency statistics
    :return: weights, complex frequencies, exact tested frequencies, estimate of residual noise and snr at those
    frequencies
    """
    epochs = set_default_unit(epochs, u.uV)
    fs = set_default_unit(fs, u.Hz)
    frequencies = set_default_unit(frequencies, u.Hz)
    delta_frequency = set_default_unit(delta_frequency, u.Hz)
    freq_axis = np.fft.rfftfreq(epochs.shape[0], 1 / fs)
    power_line_frequencies = power_line_frequency * np.arange(1,
                                                              np.round(0.5 * fs / power_line_frequency))
    _power_idx = np.array([np.argmin(np.abs(freq_axis - _freq)) for _freq in power_line_frequencies])
    _target_idx = np.array([np.argmin(np.abs(freq_axis - _freq)) for _freq in frequencies])
    # we always keep the input frequencies, even if they are multiple of line_power
    _power_idx = np.setdiff1d(_power_idx, _target_idx).astype(int)
    # get all frequencies for the analysis
    _idx_to_include_all = np.array([])
    for _idx_t, _f in enumerate(frequencies):
        # _idx_target = np.argmin(np.abs(freq_axis - _f))
        # _f_target = freq_axis[_idx_target]
        _ini_bin = np.maximum(np.argmin(np.abs(freq_axis - (_f - delta_frequency))), 0)
        _end_bin = np.minimum(np.argmin(np.abs(freq_axis - (_f + delta_frequency))), freq_axis.size - 1)
        _idx_freq = np.arange(_ini_bin, _end_bin + 1)
        _idx_to_include_all = np.concatenate((_idx_to_include_all, _idx_freq))
    _idx_to_include_all = np.unique(_idx_to_include_all)
    # remove power line bins
    _idx_to_include = np.setdiff1d(_idx_to_include_all, _power_idx).astype(int)
    _idx_excluded = np.setdiff1d(_idx_to_include, _idx_to_include_all)
    if _idx_excluded.size > 0:
        print('Excluding {:} to avoid power line harmonics'.format(
            freq_axis[_idx_excluded]))
    # print('Test frequency {:}. Noise estimation between {:} and {:}'.format(_f_target,
    #                                                                         freq_axis[_ini_bin],
    #                                                                         freq_axis[_end_bin]))

    # get target and surrounding frequency bins first
    _w, _amp, _, _rn, _snr, _exact_frequencies, y_fft = get_discrete_frequencies_weights(
        epochs=epochs,
        fs=fs,
        block_size=block_size,
        frequencies=freq_axis[_idx_to_include],
        weighted_average=weighted_average
    )
    _signal_plus_noise_var = _amp ** 2.0
    # pool across frequencies
    _var_ave = 1 / _w
    total_variance = np.mean(_var_ave, axis=0, keepdims=True)
    weights = 1 / total_variance
    # ensure same size as input epochs. All time examples are weighted the same
    weights = np.ones(epochs.shape) * weights
    pooled_rn = np.sqrt(np.mean(_rn ** 2.0, axis=0))
    pooled_snr = np.mean(_signal_plus_noise_var, axis=0) / (pooled_rn ** 2) - 1
    _idx_targets = np.array([np.argmin(np.abs(_exact_frequencies - _f)) for _f in frequencies])
    by_freq_rn = _rn[_idx_targets]
    freq_samples = y_fft[_idx_targets]
    exact_frequencies = _exact_frequencies[_idx_targets]
    by_freq_snr = _snr[_idx_targets]

    print("Weights, noise, and snr estimated by pooling {:} individual measures, each estimated with a delta of "
          "{:} around each frequency.".format(frequencies, delta_frequency))
    return weights, pooled_rn, pooled_snr, by_freq_rn, by_freq_snr, freq_samples, exact_frequencies



from peegy.processing.tools.epochs_processing_tools import et_subtract_correlated_ref, filt_data
from peegy.processing.tools.template_generator.auditory_waveforms import eog
import numpy as np
import matplotlib.pyplot as plt
import numpy.matlib

n_ch = 10
fs = 16384.0
recording_time = 16
eog_rate = 1 / 4
time = np.arange(0, 4, 1/fs)
n_trials = 10
# we generate a fake low frequency artefact signal and generate a mixing matrix
time = np.arange(0, recording_time, 1/fs)
n_samples = time.size
time_artefact = np.arange(0, 1/3, 1/fs)
n_events = int(recording_time // (1 / eog_rate))
# we generate a fake low frequency artefact signal and generate a mixing matrix
eog1_template, _ = eog(fs)
eog1mix = np.zeros((n_samples, n_ch, n_trials))
for _t in np.arange(n_trials):
    events = np.zeros(n_samples, )
    _intervals = 1 / eog_rate + 0.5 * np.random.randn(n_events, )
    event_times = np.cumsum(_intervals)
    event_times = event_times[event_times < recording_time]
    _idx = np.minimum(event_times * fs, n_samples - 1).astype(int)
    events[_idx] = 1 + 0.5 * np.random.rand(_idx.size)
    eog1 = filt_data(eog1_template.reshape(-1, 1), events.flatten(), mode='full').flatten()
    eog1 = eog1[0: n_samples]
    w1 = np.random.rand(1, n_ch)
    w1[0, -1] = 1
    eog1mix[:, :, _t] = np.matlib.repmat(eog1, n_ch, 1).T * w1

# generate EEG data at a higher frequency
ep = np.sin(2 * np.pi * 137.0 * time)
w3 = np.random.rand(1, n_ch)
epochs = np.matlib.repmat(ep, n_ch, 1).T * w3
epochs = np.expand_dims(epochs, axis=2) * np.ones((ep.size, n_ch, n_trials))
epochs[:, -1, :] = 0

# generate noise and mixing matrix
noise = 0.5 * np.random.randn(n_samples, n_ch, n_trials)
data = epochs + noise + eog1mix
orig = data.copy()
# remove artefacts
clean_data = et_subtract_correlated_ref(data=data, idx_ref=np.array([n_ch - 1]), fs=fs)

# get frequency responses
freq = np.arange(0, time.size) * fs / time.size
yfft_data = np.abs(np.fft.fft(orig, axis=0))
yfft_clean = np.abs(np.fft.fft(clean_data, axis=0))

fig = plt.figure()
ax = fig.add_subplot(221)
ax.plot(time, np.squeeze(orig[:, 0, :]), )
ax.set_title('contaminated')
ax = fig.add_subplot(222)
ax.plot(freq, np.squeeze(yfft_data[:, 0, :]))
ax.set_title('contaminated fft')
ax.set_xlim(0, 200)
ax = fig.add_subplot(223)
ax.plot(np.squeeze(clean_data[:, 0, :]))
ax.set_title('clean')
ax = fig.add_subplot(224)
ax.plot(freq, np.squeeze(yfft_clean[:, 0, :]))
ax.set_title('clean fft')
ax.set_xlim(0, 200)
fig.legend()
fig.show()
plt.show()



from peegy.definitions import edf_bdf_reader as br
import numpy as np
import matplotlib.pyplot as plt
# import bids.layout
import os
import gc
__author__ = 'jundurraga'

test_frequencies = np.array([6.8, 20.4, 40.9])
_path = os.path.abspath(os.path.dirname(__file__))
folder_name = os.path.join(_path, "../test_data/set_1/")
# layout = bids.layout.BIDSLayout(folder_name)
# f_name = layout.get(type='eeg', extensions='bdf')[0].filename
f_name = ''
reader = br.EdfBdfDataReader(file_name=f_name, channels_to_process=list(range(2)), layout_setup='biosemi64_2_EXT.lay',
                             delete_all=True)
triggers = reader.get_triggers(trigger_codes=[])
w_triggers = reader.merge_triggers(triggers, trigger_code=int('00100000', 2))
n_triggers = reader.merge_triggers(triggers, trigger_code=int('01000000', 2))
c_triggers = reader.merge_triggers(triggers, trigger_code=int('00000010', 2))
print(len(w_triggers['idx']), len(n_triggers['idx']), len(c_triggers['idx']))
reader.set_epochs_raw(triggers=triggers)
reader.set_epochs_raw_ave()

fig = plt.figure(num=1, dpi=100, facecolor='w', edgecolor='w')
fig.set_size_inches(10, 7)
ax = fig.add_subplot(111)
for i in range(30):
    ax.plot(reader.epochs_raw_ave.average)

plt.close(fig)
del reader
gc.collect()



import copy
from peegy.definitions.channel_definitions import Domain, ChannelItem, ChannelType
from peegy.processing.tools.epoch_tools.transform import de_epoch
from peegy.definitions import tables
from peegy.directories.tools import DirectoryPaths
from peegy.definitions.events import Events, SingleEvent
from peegy.tools.units.unit_tools import set_default_unit
from peegy.io.exporters import wav_writer
import inspect
import tempfile
import base64
from PIL import Image
import pandas as pd
import io
import gc
import astropy.units as u
import abc
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
import pyqtgraph as pg
import pyqtgraph.exporters
import time
from typing import List
pg.setConfigOption('leftButtonPan', False)


class DataNode(object):
    """
    Core data object used to keep data. During the creation, it requires the sampling rate and the actual data.
    Data is passed in a numpy array of n * m or n * m * t, where n represent samples, m channels, and t trials.
    Units are passed via unit kargs (default units is microVolts)
    """
    def __init__(self,
                 fs=None,
                 data: np.array = np.array([]),
                 domain=Domain.time,
                 x_offset=0.0 * u.s,
                 n_fft=None,
                 layout: np.array([ChannelItem]) = None,
                 process_history: List = [],
                 events: Events | None = None,
                 events_annotations: () = None,
                 paths: DirectoryPaths = DirectoryPaths(),
                 rn: type(np.array) | None = None,
                 rn_df: type(np.array) | None = None,
                 cum_rn: type(np.array) | None = None,
                 snr: type(np.array) | None = None,  # it may content snr for different ROIs
                 cum_snr: type(np.array) | None = None,
                 s_var=None,
                 alpha=0.05,
                 n: int | None = None,
                 w: type(np.array) | None = None,
                 # statistical_tests: tables.Tables = tables.Tables(),
                 processing_tables_global: tables.Tables = tables.Tables(),
                 # processing_tables_local: tables.Tables = tables.Tables(),
                 peaks=None,
                 # peak_frequency=None,
                 # peak_to_peak_amplitudes=None,
                 peak_time_windows=None,
                 roi_windows=None,
                 markers=None,
                 frequency_resolution=None
                 ):
        # n * m * t data array
        # if data represent time or frequency
        # time offset from origin
        # layout
        # list used to store processing history that resulted in the current data

        self._data = set_default_unit(data, u.dimensionless_unscaled)
        self.fs = fs
        self.domain = domain
        self.x_offset = x_offset
        self.layout = layout
        self.events = events
        self.events_annotations = events_annotations
        self.process_history = process_history
        self.paths = paths
        # signal statistics
        self.rn = rn
        self.cum_rn = cum_rn
        self.snr = snr
        self.cum_snr = cum_snr
        self.s_var = s_var
        self.n = n
        self.rn_df = rn_df
        self.w = w
        self.alpha = alpha
        # statistical data
        self._statistical_tests = tables.Tables()
        # other processing tables
        self._processing_tables_global = processing_tables_global
        self._processing_tables_local = tables.Tables()
        # peak measures
        self.peaks = peaks
        self.peak_time_windows = peak_time_windows
        self.roi_windows = roi_windows
        self.markers = markers
        self.n_fft = n_fft
        self.frequency_resolution = frequency_resolution
        if self.layout is None and data.size:
            self.layout = np.array([ChannelItem() for _ in range(self.data.shape[1])])  # default channel information

        self._x = None
        self._y = None

    def get_x(self):
        if self.domain == Domain.time:
            out = np.arange(self.data.shape[0]) / self.fs - self.x_offset
            out = out.to(u.s)
        if self.domain == Domain.frequency:
            out = np.arange(self.data.shape[0]) * self.fs / self.n_fft
            out = out.to(u.Hz)
        if self.domain == Domain.time_frequency:
            out = self._x
        return out

    def set_x(self, value):
        self._x = value
    x = property(get_x, set_x)

    def get_y(self):
        return self._y

    def set_y(self, value):
        self._y = value

    y = property(get_y, set_y)

    def get_processing_tables_global(self):
        return self._processing_tables_global

    def set_processing_tables_global(self, value: tables.Tables() = None):
        self._processing_tables_global.append(value)

    processing_tables_global = property(get_processing_tables_global, set_processing_tables_global)

    def get_processing_tables_local(self):
        return self._processing_tables_local

    def set_processing_tables_local(self, value: tables.Tables() = None):
        self._processing_tables_local.append(value)

    processing_tables_local = property(get_processing_tables_local, set_processing_tables_local)

    def get_statistical_tests(self):
        return self._statistical_tests

    def set_get_statistical_tests(self, value: tables.Tables() = None):
        self._statistical_tests.append(value)

    statistical_tests = property(get_statistical_tests, set_get_statistical_tests)

    def apply_layout(self, layout=None):
        for _s_i in layout.layout:
            for _i, _l in enumerate(self.layout):
                if _l.idx == _s_i.idx:
                    self.layout[_i] = _s_i
                    break

    def delete_channel_by_idx(self, idx: np.array = np.array([])):
        labels = '/'.join([_ch.label for _ch in self.layout[idx]])
        self.data = np.delete(self.data, idx, 1)
        self.layout = np.delete(self.layout, idx)
        print('removed channels: ' + ''.join(labels))

    def delete_channel_by_label(self, label=''):
        _idx = np.array([_i for _i, _lay in enumerate(self.layout) if _lay.label == label])
        if _idx.size:
            self.delete_channel_by_idx(idx=_idx)

    def get_channel_idx_by_label(self, labels: np.array([str]) = None):
        if labels is None:
            labels = np.array([_channel.label for _channel in self.layout])
        if isinstance(labels, list):
            labels = np.array(labels)
        # ensure unique labels and remove repeated in same order
        _, idx_labels = np.unique(labels, return_index=True)
        labels = labels[np.sort(idx_labels)]
        all_labels = np.array([_channel.label for _idx, _channel in enumerate(self.layout)])
        idx = []
        for _label in labels:
            _idx = np.argwhere(_label == all_labels)
            if _idx.size:
                idx.append(_idx.squeeze())
        idx = np.array(idx)
        return idx

    def get_channel_idx_by_type(self, channel_type=ChannelType.Event):
        all_idx = np.array([_idx for _idx, _channel in enumerate(self.layout) if _channel.type == channel_type])
        idx = np.unique(all_idx)
        return idx

    def get_channel_label_by_idx(self, idx: np.array = np.array([])) -> [str]:
        if idx.size == 0:
            idx = np.arange(self.layout.size)
        labels = [_ch.label for _ch in self.layout[idx]]
        return labels

    def set_data(self, value=np.array):
        self._data = value

    def get_data(self):
        return self._data
    data = property(get_data, set_data)

    def data_slice(self, ini_sample: int = 0, end_sample: int = np.inf, axis: int = 0):
        # obtain data from ini_sample until but not including the end_sample
        ini_sample = np.maximum(ini_sample, 0)
        end_sample = np.minimum(self._data.shape[axis], end_sample)
        indices = np.arange(ini_sample, end_sample).astype(int)
        out = np.take(self._data, indices=indices, axis=axis)
        return out

    def get_max_snr_per_channel(self):
        if self.snr is not None:
            max_snr = np.nanmax(np.atleast_2d(self.snr), axis=0)
        else:
            max_snr = np.array([None] * self._data.shape[1])
        return np.atleast_1d(np.squeeze(max_snr))

    def get_max_s_var_per_channel(self):
        max_var = np.nanmax(self.s_var, axis=0)
        return np.atleast_1d(np.squeeze(max_var))

    def x_to_samples(self,
                     value: type(np.array) | None = None,
                     ) -> np.array:
        out = np.array([], dtype=int)
        if self.domain == Domain.time:
            _offset = self.x_offset
            _scaling = self.fs
        if self.domain == Domain.frequency:
            _offset = 0 * u.Hz
            _scaling = 1 / self.frequency_resolution
        for _v in value:
            if np.isinf(_v):
                out = np.append(out, self.data.shape[0] - 1)
            else:
                # out = np.append(out, np.argmin(np.abs(self.x.value - np.array(_value)))).astype(int)
                sample = np.round(((_v + _offset) * _scaling).to(u.dimensionless_unscaled))
                sample = np.minimum(sample, self.data.shape[0] - 1)
                sample = np.maximum(sample, 0)
                out = np.append(out, int(sample.value))
        return out

    def x_range_to_samples(self,
                           ini_time: type(u.Quantity) | None = None,
                           end_time: type(u.Quantity) | None = None,
                           ) -> np.array:
        if ini_time is None:
            ini_time = self.x[0]
        if end_time is None:
            end_time = self.x[-1]
        ini_sample = self.x_to_samples([ini_time])
        end_sample = self.x_to_samples([end_time])
        out = np.arange(ini_sample, end_sample + 1).astype(int)
        return out

    def samples_to_x(self, value: np.array(int)) -> np.array(u.Quantity):
        out = np.array([]) * 1 / self.fs.unit
        for _v in value:
            out = np.append(out, _v / self.fs - self.x_offset)
        return out

    def data_to_pandas(self):
        _row_test = []
        if self._data is not None and self._data.size:
            for i in range(self._data.shape[1]):
                data_binary_x = io.BytesIO()
                np.save(data_binary_x, self.x.value)
                data_binary_y = io.BytesIO()
                np.save(data_binary_y, self.data[:, i].value)
                data_binary_x.seek(0)
                data_binary_y.seek(0)
                _row_time_data = {'domain': self.domain,
                                  'channel': self.layout[i].label,
                                  'fs': self.fs,
                                  'x': data_binary_x.read(),
                                  'y': data_binary_y.read(),
                                  'x_unit': self.x.unit.to_string(),
                                  'y_unit': self.data[:, i].unit.to_string(),
                                  'snr': self.get_max_snr_per_channel()[i]}
                _row_test.append(_row_time_data)
        _data_pd = pd.DataFrame(_row_test)
        return _data_pd

    def append_new_channel(self, new_data: type(np.array) | None = None, layout_label: str = ''):
        new_channel = ChannelItem(label=layout_label)
        self.layout = np.append(self.layout, new_channel)
        self.data = np.append(self.data.value, new_data.to_value(self.data.unit), axis=1) * self.data.unit
        self.rn = np.append(self.rn, np.nan)
        if self.snr is not None:
            self.snr = np.append(self.snr, np.nan)
        if self.s_var is not None:
            self.s_var = np.append(self.s_var, np.nan)

    def clip(self, ini_time: u.Quantity = 0 * u.s, end_time: u.Quantity = np.inf * u.s):
        ini_time = set_default_unit(ini_time, u.s)
        end_time = set_default_unit(end_time, u.s)
        if np.isinf(end_time):
            end_time = self.x[-1]
        samples = self.x_to_samples(np.array([ini_time.to(u.s).value,
                                              end_time.to(u.s).value]) * u.s)
        _extact_time_offset = self.x[samples[0]]
        self._data = self._data[samples[0]: samples[-1], ::]
        self.events.clip(ini_time=ini_time, end_time=end_time)
        # apply offset
        self.events.set_offset(time_offset=_extact_time_offset)

    def randomize_events(self, epoch_length: type(u.Quantity) | None = None,
                         n_events: int | None = None,
                         event_code: float = 1,
                         dur: u.Quantity = 100 * u.us,
                         min_time: type(u.Quantity) | None = None,
                         max_time: type(u.Quantity) | None = None):
        """
        Generate random time events.
        :param epoch_length: Time of epoch length. This value will only affect the maximum timing of the random events
        so that when the data is epoched with the desired epoch length the last event position will not be beyond the
        total timing of the data minus the epoch length. In this way ensuring that all random event will be used when
        epoching the data.
        :param n_events: Number of random events to be generated. If empty the original event number is used
        :param event_code: float with the event code that will be used to generate/assign new random events
        :param dur: time duration that will be assigned to the new random events
        :param min_time: minimum time of new events
        :param max_time: maximum time of new events
        :return: None
        """
        # generate random events based on original number of events or given number of events
        _n_events = None
        _min_time, _max_time = None, None
        if self.events.get_events_code(code=event_code).size:
            _n_events = self.events.get_events_code(code=event_code).size
            _min_time = self.events.get_events_time(code=event_code).min()
            _max_time = self.events.get_events_time(code=event_code).max()
        if n_events is None:
            n_events = _n_events
        if min_time is not None:
            _min_time = min_time

        if max_time is not None:
            _max_time = max_time

        if _min_time is None:
            _min_time = 0 * u.s
        if _max_time is None:
            _max_time = self._data.shape[0] / self.fs

        if n_events:
            if epoch_length is None:
                epoch_length = ((_max_time - _min_time) / n_events)
            new_event_times = np.sort(
                _min_time + np.random.rand(n_events) * (_max_time - _min_time - epoch_length), kind='quicksort')
            events = np.empty(new_event_times.size, dtype=SingleEvent)
            for i, _s, in enumerate(new_event_times):
                events[i] = SingleEvent(code=event_code,
                                        time_pos=_s,
                                        dur=dur)
            self.events = Events(events=np.array(events))


class InputOutputQtFigure(object):
    """
    This class is a wrapper to allow sphinx showing qt figures
    """
    def __init__(self, figure: pg.GraphicsView | None = None):
        self.figure = figure

    def _repr_html_(self):
        # self.figure.scene().setSceneRect(0, 0, 600, 300)
        # ex = pg.exporters.SVGExporter(self.figure.scene())
        _file = tempfile.NamedTemporaryFile(delete=False,
                                            suffix='.png')
        ex = pg.exporters.ImageExporter(self.figure.scene())
        target_width_inches = 6
        target_dpi = 300
        ex.parameters()['width'] = target_width_inches * target_dpi
        # ex.parameters()['height'] = 300
        ex.export(_file.name)
        # html__figure_exporter = out.decode("utf-8")
        img = Image.open(_file.name)
        byte_arr = io.BytesIO()
        img.save(byte_arr, format='PNG')  # Save image to the buffer
        png_data = base64.b64encode(byte_arr.getvalue()).decode('utf-8')
        out = f'<img src="data:image/png;base64, {png_data}"/>'
        return out


class InputOutputProcess(metaclass=abc.ABCMeta):
    """
    Core abstract input-output function. Each inherit class must provide an input data, a process function and and
    output data set
    """
    def __init__(self, input_process: object | None = None,
                 keep_input_node=True,
                 **kwargs):
        self.input_process: InputOutputProcess = input_process if input_process is not None else self  # Input data
        # must be InputOutputProcess class
        self.function_args: dict = kwargs
        self.input_node: DataNode | None = None
        self.output_node: DataNode | None = None
        self._process_parameters: dict = kwargs  # dict with all the parameters used in process function
        self._keep_input_node: bool = keep_input_node
        self._figures: List[plt.Figure] | None = None
        self._name: str | None = None
        self._ready = False
        if self._name is None:
            self._name = self.__class__.__name__
        self._calling_parameters = None
        self._figures_path = None
        if 'calling_parameters' in kwargs.keys():
            self._calling_parameters = kwargs['calling_parameters']

    def get_process_parameters(self):
        return self._process_parameters

    def set_process_parameters(self, value):
        self._process_parameters = value
    process_parameters = property(get_process_parameters, set_process_parameters)

    def get_keep_input_node(self):
        return self._keep_input_node

    def set_keep_input_node(self, value):
        self._keep_input_node = value
    keep_input_node = property(get_keep_input_node, set_keep_input_node)

    def get_figures(self):
        return self._figures

    def set_figures(self, value):
        self._figures = value
    figures = property(get_figures, set_figures)

    def get_figures_path(self):
        return self._figures_path

    def set_figures_path(self, value):
        self._figures_path = value
    figures_path = property(get_figures_path, set_figures_path)

    def get_name(self):
        return self._name

    def set_name(self, value):
        self._name = value
    name = property(get_name, set_name)

    def get_ready(self):
        return self._ready

    def set_ready(self, value):
        self._ready = value
    ready = property(get_ready, set_ready)

    def transform_data(self):
        return 'core method to generate output data'

    def run(self):
        if (isinstance(self.input_process, InputOutputProcess) and self != self.input_process and
                not self.input_process.ready):
            self.input_process.run()
        self.input_node = self.input_process.output_node
        callable_list = [_i for _i in dir(DataNode) if callable(getattr(DataNode, _i))]
        properties_list = [_i[0] for _i in inspect.getmembers(DataNode, lambda o: isinstance(o, property))]
        attributes = [a for a in dir(self.input_node) if not (a.startswith('__') or a.startswith('_')) and
                      a not in callable_list and a not in properties_list]
        pars = {_a: getattr(self.input_node, _a) for _a in attributes}
        self.output_node = DataNode(data=np.array([]),
                                    **pars
                                    )  # Output data must be DataNode class
        start = time.time()
        self.transform_data(**self.function_args)
        end = time.time()
        print('Elapsed time - {:}: {:.3f}s'.format(self.name, end - start))
        self.ready = True

        if not self.keep_input_node:
            if isinstance(self.input_node, DataNode):
                self.input_node.data = None
            self.input_node = None
            gc.collect()
        return

    def plot(self,
             plot_input: bool = False,
             plot_output: bool = True,
             ch_to_plot: List[str] | None = None,
             interactive: bool = True,
             show_events: bool = True,
             channels_offset: type(u.Quantity) | None = None):
        """
        This method will plot data in the input and/or the output node.
        :param plot_input: bool indicating if input_node data should be plotted
        :param plot_output: bool indicating if output_node data should be plotted
        :param ch_to_plot: list of strings indicating the labels of the channels to be shown. If empty, all channels
        will be shown.
        :param interactive: boolean indicating if plot should be interactive or not
        :param show_events: if True, events will be shown
        :param channels_offset: offset in vertical units to separate channels in figure
        """
        win = plot_input_output_process(input_output_process=self,
                                        plot_input=plot_input,
                                        plot_output=plot_output,
                                        ch_to_plot=ch_to_plot,
                                        channels_offset=channels_offset,
                                        show_events=show_events
                                        )
        out = InputOutputQtFigure(figure=win)
        if interactive:
            win.show()
            pg.QtGui.QGuiApplication.instance().exec_()
        return out

    def get_input_parameters(self):
        all_input_variables = list(inspect.signature(self.__init__).parameters.keys())
        # all_locals = self.__init__.__code__.co_varnames
        # attributes = [a for a in dir(self.input_node) if not (a.startswith('__') or a.startswith('_')) and
        #               a not in callable_list and a not in properties_list]
        parameters = {key: self.__dict__[key] for key in all_input_variables if key in self.__dict__.keys()}

        return parameters

    def to_wav(self,
               file_name: str | None = None,
               channel_labels: List[str] | None = None,
               fs: u.Quantity = 48000 * u.Hz,
               event_code: float | None = None,
               event_duration: type(u.Quantity) | None = None,
               normalize: bool = False,
               demean_edges=True,
               edge_duration: type(u.Quantity) | None = None,
               gain: float = 0.99,
               ):
        _idx_ch = self.output_node.get_channel_idx_by_label(labels=channel_labels)
        if _idx_ch.size:
            selected_data = self.output_node.data[:, _idx_ch, ...]
            if np.ndim(selected_data) == 2:
                _events = Events(events=self.output_node.events.get_events(code=event_code))
                _data = selected_data
            if np.ndim(selected_data) == 3:
                _data, _events = de_epoch(data=selected_data,
                                          demean_edges=demean_edges,
                                          edge_duration=edge_duration,
                                          event_code=event_code,
                                          event_duration=event_duration,
                                          fs=self.input_node.fs)
            wav_writer.data_to_wav(data=_data,
                                   fs=self.output_node.fs,
                                   fs_wav=fs,
                                   events=_events,
                                   output_file_name=file_name,
                                   normalize=normalize,
                                   gain=gain)


def plot_input_output_process(input_output_process: InputOutputProcess | None = None,
                              plot_input: bool = False,
                              plot_output: bool = True,
                              ch_to_plot: List[str] | None = None,
                              de_mean: bool = False,
                              show_events: bool = True,
                              channels_offset: type(u.Quantity) | None = None):
    """
    This method will plot data in the input and/or the output node.
    :param input_output_process: InputOutputProcess for which plot will be generated.
    :param plot_input: bool indicating if input_node data should be plotted
    :param plot_output: bool indicating if output_node data should be plotted
    :param ch_to_plot: list of strings indicating the labels of the channels to be shown. If empty, all channels
    will be shown.
    :param de_mean: if True, lines will be centred around the mean and then offset
    :param show_events: if True, events will be shown
    :param channels_offset: offset in vertical units to separate channels in figure
    """
    if not plot_input and not plot_output:
        return
    data_in = None
    if input_output_process.input_node is not None:
        data_in = copy.copy(input_output_process.input_node.data)
        name_in = input_output_process.input_process.name
        if input_output_process.input_node.domain == Domain.frequency:
            data_in = np.abs(data_in)
    data_out = None
    if input_output_process.output_node is not None:
        name_out = input_output_process.name
        data_out = copy.copy(input_output_process.output_node.data)
        if input_output_process.output_node.domain == Domain.frequency:
            data_out = np.abs(data_out)

    if data_in is not None or data_out is not None:
        win = pg.GraphicsLayoutWidget()
        win.setWindowTitle(input_output_process.name + '/' + input_output_process.name)
        ax = win.addPlot(row=1, col=1)
        ax.addLegend()
    else:
        return

    offset_in, offset_out = 0, 0
    channels_in, channels_out = [], []
    if data_in is not None and plot_input:
        offset_in = (np.max(np.abs(data_in.flatten() - np.mean(data_in.flatten()))) +
                     np.min(np.abs(data_in.flatten() - np.mean(data_in.flatten())))) / 2
        channels_in = [_l.label for _l in input_output_process.input_node.layout]
    if data_out is not None and plot_output:
        offset_out = (np.max(np.abs(data_out.flatten() - np.mean(data_out.flatten()))) +
                      np.min(np.abs(data_out.flatten() - np.mean(data_out.flatten())))) / 2
        channels_out = [_l.label for _l in input_output_process.output_node.layout]

    if ch_to_plot is not None:
        channels_in = list(set(ch_to_plot) & set(channels_in))
        channels_out = list(set(ch_to_plot) & set(channels_out))

    channels = np.unique(channels_in + channels_out)
    if channels_offset is None:
        offset = np.maximum(offset_in, offset_out)
    else:
        offset = channels_offset
    if plot_input and data_in is not None:
        x_in = input_output_process.input_node.x
        ax.setDownsampling(auto=True, mode='peak')
        ax.showGrid(True, True)
        if de_mean:
            data_in = data_in - np.mean(data_in, axis=0)
        _color = (255, 0, 0)
        for _i, _label in enumerate(tqdm(channels_in, desc='Generating input plot')):
            _name = name_in if _i == 0 else None
            _pos = np.argwhere(channels == _label).squeeze()
            if not _pos.size:
                continue
            text_item = pg.TextItem(_label, anchor=(0.0, 0.0))
            text_item.setPos(0, offset.value * _pos)
            if data_in.ndim == 2:
                ax.plot(x_in, data_in[:, _i] +
                        offset * _pos, pen=_color, name=_name)

            if data_in.ndim == 3:
                data_in[:, _i, :] = data_in[:, _i, :] + offset * _pos
                for _t in tqdm(range(data_in.shape[2]),
                               desc='Generating input plot per epochs for channel {:}'.format(_label)):
                    ax.plot(x_in, data_in[:, _i, _t], pen=_color, name=_name if _t == 0 else None)
                    pg.QtCore.QCoreApplication.processEvents()
            ax.addItem(text_item)

            if show_events and input_output_process.input_node.events is not None:
                _codes = np.unique(input_output_process.input_node.events.get_events_code(include_bad_events=True))
                for _code in _codes:
                    _events = input_output_process.input_node.events.get_events(code=_code, include_bad_events=True)
                    _good_events = np.array([])
                    _bad_events = np.array([])
                    for _ev in _events:
                        if _ev.bad_event:
                            _bad_events = np.append(_bad_events, _ev.time_pos.value)
                        else:
                            _good_events = np.append(_good_events, _ev.time_pos.value)
                    if _good_events.size:
                        _pvl = plotVerticalLines()
                        ax.addItem(_pvl)
                        _pvl.setVerticalLines(_good_events, pen=_color)
                    if _bad_events.size:
                        _pvl = plotVerticalLines()
                        ax.addItem(_pvl)
                        _pvl.setVerticalLines(_bad_events, pen='r')

        if input_output_process.input_node.domain == Domain.time:
            ax.setLabel('bottom', "Time [{:}]".format(input_output_process.input_node.x.unit))
            ax.setLabel('left', "Amplitude [{:}]".format(input_output_process.input_node.data.unit))
        if input_output_process.input_node.domain == Domain.frequency:
            ax.setLabel('bottom', "Frequency [{:}]".format(input_output_process.input_node.x.unit))
            ax.setLabel('left', "Amplitude [{:}]".format(input_output_process.input_node.data.unit))

    if plot_output and data_out is not None:
        _color = (0, 0, 255)
        x_out = input_output_process.output_node.x
        ax.setDownsampling(ds=100, mode='peak')
        ax.showGrid(True, True)
        if de_mean:
            data_out = data_out - np.mean(data_out, axis=0)
        for _i, _label in enumerate(tqdm(channels_out, desc='Generating output plot')):
            _name = name_out if _i == 0 else None
            _pos = np.argwhere(channels == _label).squeeze()
            if not _pos.size:
                continue
            text_item = pg.TextItem(_label, anchor=(0.0, 0.0))
            text_item.setPos(0, offset.value * _pos)
            if data_out.ndim == 2:
                ax.plot(x_out, data_out[:, _i] + offset * _pos, pen=_color, name=_name)
            if data_out.ndim == 3:
                data_out[:, _i, :] = data_out[:, _i, :] + offset * _pos
                for _t in tqdm(range(data_out.shape[2]),
                               desc='Generating output plot per epochs for channel {:}'.format(_label)):
                    ax.plot(x_out, data_out[:, _i, _t], pen=_color, name=_name if _t == 0 else None)
                    pg.QtCore.QCoreApplication.processEvents()

            ax.addItem(text_item)
            if show_events and input_output_process.output_node.events is not None:
                _codes = np.unique(input_output_process.output_node.events.get_events_code())
                for _code in _codes:
                    _events = input_output_process.output_node.events.get_events(code=_code, include_bad_events=True)
                    _good_events = np.array([])
                    _bad_events = np.array([])
                    for _ev in _events:
                        if _ev.bad_event:
                            _bad_events = np.append(_bad_events, _ev.time_pos.value)
                        else:
                            _good_events = np.append(_good_events, _ev.time_pos.value)
                    if _good_events.size:
                        _pvl = plotVerticalLines()
                        ax.addItem(_pvl)
                        _pvl.setVerticalLines(_good_events, pen=_color)

                    if _bad_events.size:
                        _pvl = plotVerticalLines()
                        ax.addItem(_pvl)
                        _pvl.setVerticalLines(_bad_events, pen='r')

        if input_output_process.output_node.domain == Domain.time:
            ax.setLabel('bottom', "Time [{:}]".format(input_output_process.output_node.x.unit))
            ax.setLabel('left', "Amplitude [{:}]".format(input_output_process.output_node.data.unit))
        if input_output_process.output_node.domain == Domain.frequency:
            ax.setLabel('bottom', "Frequency [{:}]".format(input_output_process.output_node.x.unit))
            ax.setLabel('left', "Amplitude [{:}]".format(input_output_process.output_node.data.unit))
    pg.QtCore.QCoreApplication.processEvents()
    return win


class plotVerticalLines(pg.PlotDataItem):
    def __init__(self, container: pg.PlotItem | None = None):
        super(plotVerticalLines, self).__init__()
        self.pen = None

    def dataBounds(self, ax, frac=1.0, orthoRange=None):
        return [None, None]

    def setVerticalLines(self, x, pen=None):
        self.pen = pen
        x = np.vstack((x, x)).T.astype(float)
        y = np.zeros_like(x)
        vb = self.getViewBox()
        xr, yr = vb.viewRange()
        y[:, 0] = yr[0]
        y[:, 1] = yr[1]
        self.setData(x.ravel(), y.ravel(), connect='pairs', pen=self.pen)
        vb.sigYRangeChanged.connect(self._y_range_changed)

    def _y_range_changed(self, viewbox):
        xr, yr = viewbox.viewRange()
        # get the original not transformed clipped etc.
        x = self.xData
        y = self.yData
        y[0::2] = yr[0]
        y[1::2] = yr[1]
        self.setData(x, y, connect='pairs', pen=self.pen)



import numpy as np


def get_local_minima(x: type(np.array) | None = None):
    _minima_idx = np.arange(x.shape[0])[np.r_[True, x[1:] < x[:-1]] & np.r_[x[:-1] < x[1:], True]]
    return x[_minima_idx], _minima_idx


def get_local_maxima(x: type(np.array) | None = None):
    _maxima_idx = np.arange(x.shape[0])[np.r_[True, x[1:] > x[:-1]] & np.r_[x[:-1] > x[1:], True]]
    return x[_maxima_idx], _maxima_idx



import matplotlib.pyplot as plt
import peegy.io.readers.generic_csv_reader as cvsr

file_path1 = r'D:\Measurements\IRU\CAP_Infants\AllData\001\20190409 12-23-21 - sub01_block1_sig0_1_65db.txt'
file_path2 = r'D:\Measurements\IRU\CAP_Infants\AllData\001\20190409 12-23-21 - sub01_block1_sig0_1_65db.txt'

header = cvsr.read_header(file_path1)
data, ev, _ = cvsr.read_channel(header=header)

plt.plot(data)
plt.plot(ev)
plt.show()
